from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from builtins import super
from builtins import int
from builtins import object
from future import standard_library
standard_library.install_aliases()
import os
from sqlite3 import OperationalError
from threading import Thread
from time import sleep
from http.server import BaseHTTPRequestHandler, HTTPServer

from .manager import INCONSISTENT, UNCERTAIN
from .args import HTTPBINDADDRESS, HTTPPORT, RETRIEVE, DAEMON, WEB
from .download import DEFAULT_NAME
from .process import ProcessManager
from .sqlite import SqliteSupport, NoResult
from .utils import process_exists, format_time_epoch, format_time_epoch_local, safe_unlink

"""
The 'rover web' command - run a web service that displays information on the download manager.
"""


class DeadMan(Thread):
    """
    Repeatedly check the parent process and exit when that dies.

    (Maybe processes should do this anyway with HUP, but this seems to be needed...)
    """

    def __init__(self, log, ppid, server, log_path):
        super().__init__()
        self._log = log
        self._ppid = ppid
        self._server = server
        self._log_path = log_path
        self._log.debug('DeadMan watching PID %d' % self._ppid)

    def run(self):
        while self._ppid != 1 and process_exists(self._ppid):
            sleep(1)
        self._log.info('Exiting because parent exited')
        if self._log_path and os.path.exists(self._log_path) and os.path.getsize(self._log_path) == 0:
                    safe_unlink(self._log_path)
        self._server.shutdown()
        sleep(1)
        exit()


class RequestHandler(BaseHTTPRequestHandler):
    """
    Generate the web page.

    BaseHTTPRequestHandler is part of the standard Python library HTTP server code - we extend it for this
    particular application.
    """

    def do_GET(self):
        """
        This method is called when the server receives a GET request.

        We detect what is running and generate the appropriate response.
        """
        self.send_response(200)
        self.send_header('Content-type', 'text/html')
        self.end_headers()
        self._html_header()
        self._write('<h1>ROVER</h1>')
        pid, command = self.server.process_manager.current_command()
        if command == DAEMON:
            self._do_daemon()
        elif command == RETRIEVE:
            self._do_retrieve()
        else:
            self._do_quiet()
        self._html_footer()

    def _write(self, text):
        self.wfile.write(text.encode('ascii'))

    def _html_header(self):
        self._write('''<html lang="en">
  <head>
    <meta charset="ascii">
    <title>ROVER</title>
    <style>
* {
  font-family: monospace;
}
    </style>
  </head>
  <body>
''')

    def _html_footer(self):
        self._write('''
  </body>
</html>
''')

    def _do_quiet(self):
        self._write('<p>No daemon or retrieve process is running.</p>')

    def _do_daemon(self):
        self._write('<h2>Subscription Status</h2>')
        count = [0]

        def callback(row):
            count[0] += 1
            (id, file, availability_url, dataselect_url, creation_epoch, last_check_epoch,
             last_error_count, consistent) = row
            self._write('<h3>Subscription %d</h3>' % id)
            self._write('''<p><pre>File: <a href="file://%s">%s</a>
Availability URL: <a href="%s">%s</a>
Dataselect URL: <a href="%s">%s</a>
Created: %s (%s local)
Last active: %s (%s local)</pre></p>''' %
                        (file, file, availability_url, availability_url, dataselect_url, dataselect_url,
                         format_time_epoch(creation_epoch), format_time_epoch_local(creation_epoch),
                         format_time_epoch(last_check_epoch) if last_check_epoch else 'never',
                         format_time_epoch_local(last_check_epoch) if last_check_epoch else 'never'
                        ))
            self._write_progress(id, last_check_epoch, last_error_count, consistent)

        try:
            self.server.foreachrow('''SELECT id, file, availability_url, dataselect_url, creation_epoch,
                                             last_check_epoch, last_error_count, consistent
                                        FROM rover_subscriptions ORDER BY id''', tuple(), callback)
        except OperationalError:
            pass
        if not count[0]:
            self._write('<p>No subscriptions</p>')
        self._write_explanation()

    def _do_retrieve(self):
        self._write('<h2>Retrieval Progress</h2>')
        self._write_progress(DEFAULT_NAME, None, None, None)
        self._write_explanation()

    def _write_progress(self, name, last_check_epoch, last_error_count, consistent):
        try:
            initial_stations, remaining_stations, initial_time, remaining_time, n_retries, download_retries = \
                self.server.fetchone('''SELECT initial_stations, remaining_stations, initial_time, remaining_time,
                                               n_retries, download_retries
                                          FROM rover_download_stats WHERE submission = ?''', (name,))
            self._write('<p>Progress for download attempt %d of %d:<pre>\n' % (n_retries, download_retries))
            self._write_bar('stations', initial_stations, remaining_stations)
            self._write_bar('timespan', initial_time, remaining_time)
            self._write('</pre></p>')
        except NoResult:
            if last_error_count:
                self._write('<p>Inactive.  WARNING: Last download had errors, so data may be incomplete.</p>')
            elif last_check_epoch:
                if consistent == INCONSISTENT:
                    self._write('''<p>Inactive.  WARNING: last download detected inconsistent web services
                                   (eg dataselect not providing data promised by availability)''')
                elif consistent == UNCERTAIN:
                    self._write('''<p>Inactive.  Last download had no errors but could not check web service consistency
                                   (unlikely to be a problem).''')
                else:
                    self._write('<p>Inactive.  Latest download had no errors.</p>')
            else:
                self._write('<p>Inactive.  Waiting for initial download.</p>')
        except OperationalError:
            self._write('<p>Error: no statistics in database.</p>')

    def _write_bar(self, label, initial, current):
        # there's some massaging of numbers here because the seconds might not match exactly
        # due to fractions of a sample being added to include boundary values
        percent = max(0, min(100, int(100 * (initial - current) / max(1, initial))))
        n = int(percent / 2 + 0.5)
        self._write('%10s: %10d/%-10d  (%3d%%)  |%s|\n' %
                    (label, initial - current, initial, percent, '#' * n + ' ' * (50-n)))

    def _write_explanation(self):
        self._write('''
<h2>Notes</h2>
<ul>
<li>Progress values are based on data still to be downloaded; they do not include data within the pipeline.</li>
<li>The stations statistic is the number of distinct Net_Sta that will be requested.</li>
<li>The timespan statistic is the total time (s) covered by the data in the downloads.</li>
<li>Firefox will not open file:// URLs, but you can copy them to the address bar, where they will work.</li>
</ul>
''')

    def log_message(self, format, *args):
        pass


class Server(HTTPServer, SqliteSupport):
    """
    Extend the standard HTTP server to include a database connection and access to process data.
    """

    def __init__(self, config, address, handler):
        HTTPServer.__init__(self, address, handler)
        SqliteSupport.__init__(self, config)
        self.process_manager = ProcessManager(config)


class ServerStarter(object):
    """
### Web

    rover web

    rover web --http-bind-address 0.0.0.0 --http-port 8080

    rover retrieve --web ...   # the default

Starts a web server that provides information on the progress of the download
manager. ROVER's default configuration starts `rover web` automatically.
The flag`--no-web` prevents ROVER's web server from launching in accordance
with `rover retrieve`.

##### Significant Options

@web
@http-bind-address
@http-port
@verbosity
@log-dir
@log-verbosity

##### Examples

    rover retrieve --no-web

will run retrieve without the web server.

    """

    def __init__(self, config):
        self._bind_address = config.arg(HTTPBINDADDRESS)
        self._http_port = config.arg(HTTPPORT)
        self._ppid = ProcessManager(config).current_command()[0]
        if not self._ppid:
            raise Exception('Cannot start web server independently of retrieve')
        self._log = config.log
        self._log_path = config.log_path
        self._config = config

    def run(self, args):
        if args:
            raise Exception('Usage: rover %s' % WEB)
        server = Server(self._config, (self._bind_address, self._http_port), RequestHandler)
        DeadMan(self._log, self._ppid, server, self._log_path).start()
        self._log.info('Starting HTTP server on http://%s:%d' % (self._bind_address, self._http_port))
        server.serve_forever()



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from builtins import super
from future import standard_library
standard_library.install_aliases()
from builtins import object
from re import sub

from .args import HELP_CMD, LIST_INDEX, DATADIR, INIT_REPOSITORY, RETRIEVE, TEMPDIR, INGEST, INDEX, SUBSCRIBE, \
    AVAILABILITYURL, DATASELECTURL, DOWNLOAD, LIST_RETRIEVE, mm, ALL, MSEEDINDEXCMD, Arguments, MDFORMAT, FILE, START, \
    STATUS, STOP, LIST_SUBSCRIBE, UNSUBSCRIBE, TRIGGER, DAEMON, LIST_SUMMARY, SUMMARY, DEFAULT_FILE, INIT_REPO, INIT, \
    RETRIEVE_METADATA, WEB
from .utils import dictionary_text_list

"""
The 'rover help' command.
"""


BACKGROUND = 'background'
USAGE = 'usage'
LOWLEVEL = 'low-level'


def usage(config):
    return '''
                    Common ROVER Commands

rover {0} [directory]

  Initializes a given directory, or the current directory
  if no argument is provided, as a ROVER data repository.
  {0} will create a configuration file, by default {0},
  as well as directories for logs and data.

  The aliases `rover {1}` and `rover {2}` also exist.

rover {3} (file | sta=... [start [end]] | N_S_L_C [start [end]])

  Compares ROVER's local index with remotely available data,
  then downloads and ingest files missing from the local
  repository. The location of remote data availability is
  determined by the config parameter {4}, and the data is
  downloaded from the URL that is configured by the parameter {5}.
  Use {7} to determine data available on the remote server which
  is not in the local repository.

rover {6} (file | sta=... [start [end]] | N_S_L_C [start [end]])

  Compares the local index with the data available remotely,
  then displays the difference. Note that the summary is
  printed to stdout, while logging is to stderr.

rover {7} ...

  List index entries for the repository, config parameter %s,
  that match the given constraints. For more information,
  run "rover {7}" with no arguments.

rover {8} ...

  List summary entries for the repository, config parameter
  {9}, that match the given constraints.  This is faster than
  `rover {10}` but gives less detail.  For more information,
  run "rover {8}" with no arguments.

rover {11} <command>

  Gives help on the various commands.

'''.format(INIT_REPOSITORY, DEFAULT_FILE, INIT,
           RETRIEVE, AVAILABILITYURL, DATASELECTURL, LIST_RETRIEVE,
           LIST_INDEX, LIST_SUMMARY, DATADIR, LIST_INDEX,
           HELP_CMD)


def background(config):
    return '''
                   Advanced ROVER Commands

rover {0}

  Start the background process that regularly downloads subscriptions.

rover {1}

rover {2}

  Display the status of, and stop, the background process.

rover {3} (file | sta=... [start [end]] | N_S_L_C [start [end]])

  Subscribe generates a background service (daemon) that regularly compares data
  available at the configured server with the local repository. If there is a
  discrepancy, available data is downloaded, ingested and indexed. `rover {3}` is
  similar to `rover {4}` but uses a daemon to regularly update the local repository.

rover {5}

rover {5} N

  Displays indices of all ROVER subscriptions. `rover {5}` is similar to
  `rover {6}`.

rover {7} N

  Ask the daemon to immediately re-process a subscription(s) based
  on the subscription's index.

rover {8} N

  Delete one or more subscriptions identified by their indices.
  Data associated with the subscription(s) is not deleted.

'''.format(START, STATUS, STOP,
           SUBSCRIBE, RETRIEVE,
           LIST_SUBSCRIBE, LIST_RETRIEVE,
           TRIGGER, UNSUBSCRIBE)


def low_level(config):
    return '''
                   Low-Level ROVER Commands

The following commands are used internally, and are often less
useful from the command line:

rover {0} url

  Downloads a single request, typically for a day, from a URL
  or a given file. File arguments are expected to contain fdsn
  web services requests and fetch data from the URL set by the
  parameter {1}. Data are downloaded to a temporary directory
  configured by the parameter {2}. After downloaded, data are
  ingested into the {3} repository and are deleted from the temp
  directory. `rover {0}` is called by {4}, {5}, and {6}.

rover {7} [(file|dir) ...]

  Indexes files, adds or changes entries in the tsindex table
  stored in the miniSEED database.

  When no argument is given, all modified files in the
  repository are processed. The `--all` flag forces
  all files to be processed. If a directory argument
  is provided, all files contained in the directory are
  processed, along with the contents of sub-directories,
  unless `--no-recurse` is specified.

rover {8} (file|dir) ...

  Adds contents from a miniSEED formatted file to ROVER's local
  repository and indexes the new data.

rover {9} ... 

Starts a web server that provides information on the progress of the download
manager. ROVER's default configuration starts `rover {9}` automatically.
The flag`--no-web` prevents ROVER's web server from launching in accordance
with `rover {4}` or `rover {10}`.

rover {11} ...

  Download missing metadata from the fdsnws-station web service and save to
  the data archive. This feature is only supported for the ASDF output format.

rover {12} ...

  Creates a summary of the index stored in a ROVER repository.
  This lists the overall span of data for each
  Net_Sta_Loc_Chan and can be queried using `rover {13}`.

'''.format(DOWNLOAD, DATASELECTURL, TEMPDIR, DATADIR, RETRIEVE, SUBSCRIBE,
           DAEMON, INDEX, INGEST, WEB, START, RETRIEVE_METADATA, SUMMARY,
           LIST_SUMMARY)


GENERAL = {
    USAGE: (usage, 'General interactive use'),
    #BACKGROUND: (background, 'Advanced use with ROVER in the background'),
    LOWLEVEL: (low_level, 'Rarely used, low-level commands')
}


class HelpFormatter(object):
    """
    Either print markdown verbatim. or strip markdown and re-justify for 80 columns.
    """

    def __init__(self, md_format):
        self._md_format = md_format

    def print_help(self, text):
        arguments = Arguments()
        first_param = True
        for line in self.__paras(text):
            if line.startswith('@'):
                if first_param:
                    arguments.print_docs_header()
                    first_param = False
                if self._md_format:
                    arguments.print_docs_row_md(line[1:])
                else:
                    arguments.print_docs_row_text(line[1:])
            elif self._md_format:
                print(self.__escape(line))
            elif line.startswith('#'):
                print(line.lstrip(' #'))
            else:
                for short in self.__splitlines(line):
                    print(short)

    def __escape(self, text):
        text = sub(r'\\', '\\\\', text)
        # text = sub(r'`', '\\`', text)
        return text

    def __paras(self, text):
        return text.splitlines()

    def __slurp(self, line):
        word = ''
        space = ''
        while line and not line[0] == ' ':
            word += line[0]
            line = line[1:]
        while line and line[0] == ' ':
            space += line[0]
            line = line[1:]
        return word, space, line

    def __splitlines(self, line):
        indentation = ''
        while line and line.startswith(' '):
            indentation += line[0]
            line = line[1:]
        if not line:
            yield ''
            return
        while line:
            short, space = indentation, ''
            line = line.lstrip()
            while line:
                word, next_space, next_line = self.__slurp(line)
                if short.strip() and len(short + space + word) > 78:
                    yield short
                    short, space = indentation, ''
                else:
                    short = short + space + word
                    space, line = next_space, next_line
            if short:
                yield short


class Helper(HelpFormatter):

    def __init__(self, config):
        super().__init__(config.arg(MDFORMAT))
        self._config = config

    def run(self, args):
        from rover import COMMANDS   # avoid import loop
        if not args:
            self._help()
            return
        elif len(args) == 1:
            command = args[0].lower()
            if command == 'help':
                self._help()
                return
            if command in COMMANDS:
                self.print_help(COMMANDS[command][0].__doc__)
                return
            elif command in GENERAL:
                self.print_help(GENERAL[command][0](self._config))
                return
        raise Exception('Help is available for: %s, %s, %s, and individual commands (or simply "rover help")' % (USAGE, BACKGROUND, LOWLEVEL))

    def _help(self):
        from rover import COMMON_COMMANDS, ADVANCED_COMMANDS   # avoid import loop
        self.print_help('''
### Help

Gives help on the various commands.

Usage:

  rover help <command>

Help is available for the following:

Core commands:
{}

General topics:
{}

Advanced commands:
{}

For example:

    rover help retrieve

'''.format(dictionary_text_list(COMMON_COMMANDS),
           dictionary_text_list(GENERAL),
           dictionary_text_list(ADVANCED_COMMANDS)))



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import super
from builtins import open
from future import standard_library
standard_library.install_aliases()
from builtins import object
from argparse import Namespace
from genericpath import exists
from os import makedirs, getcwd
from os.path import isabs, join, realpath, abspath, expanduser, dirname
from re import compile, sub

from .args import Arguments, LOGDIR, LOGSIZE, LOGCOUNT, LOGVERBOSITY, \
    VERBOSITY, LOGUNIQUE, LOGUNIQUEEXPIRE, FILEVAR, DIRVAR, TEMPDIR, DATADIR, \
    COMMAND, unbar, DYNAMIC_ARGS, INIT_REPOSITORY, m, F, FILE, FULLCONFIG, ASDF_FILENAME
from .logs import init_log, log_name
from .sqlite import init_db
from .utils import safe_unlink, canonify

"""
Package common data used in all/most classes (db connection, logs and options).
"""


class BaseConfig(object):
    """
    The configuration of the system (log, options, database).

    The Config subclass provides a different constructor.
    """

    def __init__(self, log, log_path, args, db, configdir):
        self.log = log
        self.log_path = log_path
        self._args = args
        self.db = db
        self._configdir = configdir
        self.args = args.args
        self.command = args.command
        # make arguments absolute
        self.set_args_absolute()

    def set_args_absolute(self):
        """
        Make file and directories have absolute paths
        """
        abs_args = {}
        for action in Arguments()._actions:
            name = action.dest
            if unbar(name) not in DYNAMIC_ARGS:
                if action.metavar in (DIRVAR, FILEVAR):
                    value = self.path(name)
                else:
                    value = self.arg(name)
                abs_args[name] = value
        abs_args = Namespace(**abs_args)
        self._args = abs_args
        self.args = abs_args.args

    def set_configdir(self, configdir):
        """
        Called only from initialisation, when using a new config dir.
        """
        self._configdir = configdir

    def arg(self, name, depth=0):
        """
        Look-up an arg with variable substitution.
        """
        name = sub('-', '_', name)
        if depth > 10:
            raise Exception('Circular definition involving %s' % name)
        try:
            value = getattr(self._args, name)
        except:
            raise Exception('Option %s does not exist' % name)
        while True:
            try:
                matchvar = compile(r'(.*(?:^|[^\$]))\${(\w+)}(.*)').match(value)
            except:
                # not a string variable
                break
            if matchvar:
                if matchvar.group(2) == 'CONFIGDIR':
                    inner = self._configdir
                else:
                    inner = self.arg(matchvar.group(2), depth=depth+1)
                try:
                    value = matchvar.group(1) + inner + matchvar.group(3)
                except:
                    raise Exception('String substitution only works with string parameters (%s)' % name)
            else:
                value = sub(r'\$\$', '$', value)
                break
        return value

    def path(self, name):
        """
        Paths have an implicit configdir if they are relative.
        """
        path = expanduser(self.arg(name))
        if self._configdir and not isabs(path):
            path = join(self._configdir, path)
        return realpath(abspath(path))

    def dir(self, name, create_dir=True):
        """
        Ensure the directory exists.
        """
        path = self.path(name)
        if create_dir and not exists(path):
            makedirs(path)
        return path

    def file(self, name, create_dir=True):
        """
        Ensure the enclosing directory exists.
        """
        path = self.path(name)
        dir = dirname(path)
        if create_dir and not exists(dir):
            makedirs(dir)
        return path

    def dump_log(self):
        """Only for tests"""
        pass


def timeseries_db(config):
    return join(config.dir(DATADIR), 'timeseries.sqlite')

def asdf_container(config):
    return join(config.dir(DATADIR), config.arg(ASDF_FILENAME))


class Config(BaseConfig):
    """
    An alternative constructor for BaseConfig (bootstrap from command line).
    """

    def __init__(self):
        argparse = Arguments()
        args, self.__config = argparse.parse_args()

        # there's a pile of ugliness here so that we delay error handling until we have logs.
        # see also comments in parse_args.
        self.__error = self.__config and not exists(self.__config)  # see logic in parse_args
        full_config = self.__config and not self.__error

        # Special case of initializing repository, set config directory
        configdir = None
        if args.command and args.command in (INIT_REPOSITORY):
            if not args.args:
                configdir = getcwd()
            elif len(args.args) == 1:
                configdir = args.args[0]
            else:
                raise Exception('Command %s takes at most one argument - the directory to initialise' %
                                INIT_REPOSITORY)
            configdir = canonify(configdir)

        # this is a bit ugly, but we need to use the base methods to construct the log and db
        # note that log is not used in base!
        super().__init__(None, None, args, None, dirname(self.__config) if full_config else configdir)
        self.log, self.log_path, self.__log_stream = \
            init_log(self.dir(LOGDIR) if full_config else None, self.arg(LOGSIZE), self.arg(LOGCOUNT),
                     self.arg(LOGVERBOSITY), self.arg(VERBOSITY), self.arg(COMMAND) or 'rover',
                     self.arg(LOGUNIQUE), self.arg(LOGUNIQUEEXPIRE))
        if full_config:  # if initializing, we have no database...
            self.db = init_db(timeseries_db(self), self.log)

    def lazy_validate(self):
        # allow Config() to be created first so we can log on error (see main()),
        if self.__error:
            self.log.error('You may need to configure the repository using `rover %s %s %s`).' %
                           (INIT_REPOSITORY, m(F), self.__config))
            raise Exception('Could not find configuration file (%s)' % self.__config)

    def dump_log(self):
        """
        Called only from initialization, when logging to an in-memory stream.
        """
        path, dir = log_name(self.dir(LOGDIR), self.arg(COMMAND))
        self.log.info('Dumping log to %s' % path)
        with open(path, 'w') as out:
            out.write(self.__log_stream.getvalue())


class RepoInitializer(object):
    """
### Init Repository

    rover init-repository [directory]

    rover init-repo [directory]

    rover init [directory]

Initializes a given directory, or the current directory if no argument is
provided, as a ROVER data repository. Init repository will create a
configuration file, rover.config, as well as log and data directories.

   The aliases `rover init-repo` and `rover int` also exist.

To avoid over-writing data, rover init-repo returns an error if
a rover.config file, data or log directory exist in the targeted directory.

##### Significant Options

@verbosity
@log-dir
@log-verbosity

##### Examples

    rover init-repository

will create the repository in the current directory.

    rover init-repository ~/rover

will create the repository in ~/rover

    """

    def __init__(self, config):
        self.__config = config
        self.__log = config.log
        self.__args = config._args

    def run(self, args):
        self.__check_empty()
        self.__create()

    def __check_empty(self):
        data_dir = self.__config.dir(DATADIR, create_dir=False)
        if exists(data_dir):
            raise Exception('The data directory already exists (%s)' % data_dir)
        config_file = self.__config.file(FILE, create_dir=False)
        if exists(config_file):
            raise Exception('The configuration file already exists (%s)' % config_file)
        log_dir = self.__config.dir(LOGDIR, create_dir=False)
        if exists(log_dir):
            raise Exception('The log directory already exists (%s)' % log_dir)
        # no need to check database because that's inside the data dir

    def __create(self):
        config_file = self.__config.file(FILE)
        self.__log.default('Writing new config file "%s"' % config_file)
        Arguments().write_config(config_file, self.__args, WRITE_FULL_CONFIG=self.__config.arg(FULLCONFIG))
        self.__config.dir(DATADIR)
        db = init_db(timeseries_db(self.__config), self.__log)
        db.execute('PRAGMA journal_mode=WAL')
        self.__config.dump_log()


def write_config(config, filename, **kargs):
    """
    Write a config file for sub-processes.
    """
    args = config._args
    temp_dir = config.dir(TEMPDIR)
    config_path = join(temp_dir, filename)
    safe_unlink(config_path)
    Arguments().write_config(config_path, args, **kargs)
    return config_path



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from future import standard_library
standard_library.install_aliases()
import sys
from io import StringIO
from logging import getLogger, StreamHandler, Formatter, DEBUG, addLevelName
from logging.handlers import RotatingFileHandler
from os import makedirs, getpid
from os.path import join, exists, isdir
from re import match

from .utils import clean_old_files, canonify, calc_bytes

"""
Support for logging.
"""

DEFAULT = 25   # a new logging level, between INFO and WARN


def level(n):
    '''
    Our log levels are 0-6 (silent - verbose).
    Logging levels are 50-10 (quiet - verbose)

    Ours        Logging
    0  None     60
    1  Critical 50
    2  Error    40
    3  Warning  30
    4  Default  25
    5  Info     20
    6  Debug    10

    '''
    n = max(0, min(n, 6))
    if n == 4:
        return 25
    elif n < 4:
        return 60 - 10 * n
    else:
        return 70 - 10 * n


def match_unique(name):
    return match(r'\w+_\d+\.log', name)


def log_name(log_dir, name):
    dir = canonify(log_dir)
    if not exists(dir):
        makedirs(dir)
    if not isdir(dir):
        raise Exception('"%s" is not a directory (log-dir)' % dir)
    path = join(dir, name + '.log')
    return path, dir


def init_log(log_dir, log_size, log_count, log_verbosity, verbosity, name, log_unique, log_unique_expire, stderr=None):
    """
    Create a log with two handlers.
    One handler is a rotated file, the other stderr.
    The file is for details, stderr for errors to the user.
    """

    addLevelName(DEFAULT, 'DEFAULT')

    if log_unique:
        name = '%s.%d' % (name, getpid())

    log = getLogger(name)
    log.setLevel(DEBUG)

    if log_dir:  # on initialization we have no log dir
        path, dir = log_name(log_dir, name)
        size = calc_bytes (log_size)
        count = max(min(log_count, 100), 1)
        file_handler = RotatingFileHandler(path, maxBytes=size, backupCount=count)
        stream = None
    else:
        stream = StringIO()
        file_handler = StreamHandler(stream)
        path, dir = None, None

    time_formatter = Formatter('%(levelname)-8s %(asctime)s: %(message)s')
    file_handler.setFormatter(time_formatter)
    file_handler.setLevel(level(log_verbosity))
    log.addHandler(file_handler)

    name_formatter = Formatter('%(name)s %(levelname)8s: %(message)s')
    stdout_handler = StreamHandler(stderr if stderr else sys.stderr)
    stdout_handler.setLevel(level(verbosity))
    stdout_handler.setFormatter(name_formatter)
    log.addHandler(stdout_handler)

    if dir:
        clean_old_files(dir, log_unique_expire * 60 * 60 * 24, match_unique, log)

    def get_file_handler():
        return file_handler
    
    def get_stdout_handler():
        return stdout_handler

    log.get_file_handler = get_file_handler
    log.get_stdout_handler = get_stdout_handler

    # monkey patch logger (LogAdapter works in 3, but fails on 2)
    def default(msg, *args, **kwargs):
        log.log(DEFAULT, msg, *args, **kwargs)

    log.default = default

    if dir:
        clean_old_files(dir, log_unique_expire * 60 * 60 * 24, match_unique, log)

    return log, path, stream


class LoggingContext(object):
    """
    Use this context manager to temporarily change the logging configuration
    and revert it back after doing something.
    """

    def __init__(self, logger, level=None, handler=None, close=True):
        self.logger = logger
        self.level = level
        self.handler = handler
        self.close = close

    def __enter__(self):
        if self.level is not None:
            self.old_level = self.logger.level
            self.logger.setLevel(self.level)
        if self.handler:
            self.old_handlers = self.logger.handlers
            self.logger.handlers = []
            self.logger.addHandler(self.handler)

    def __exit__(self, et, ev, tb):
        if self.level is not None:
            self.logger.setLevel(self.old_level)
        if self.handler:
            self.logger.handlers = self.old_handlers
        if self.handler and self.close:
            self.logger.handlers = self.old_handlers
            self.handler.close()
        # implicit return of None => don't swallow exceptions



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from future import standard_library
standard_library.install_aliases()
from rover import main

# this is the command executed when the module is invoked
# using 'python -m rover'

main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import super
from builtins import next
from future import standard_library
standard_library.install_aliases()
from builtins import object
from genericpath import isdir
from os import listdir, makedirs
from os.path import split, join, isfile, exists, getmtime
from sqlite3 import OperationalError

from .args import DATADIR, ALL, RECURSE
from .sqlite import SqliteSupport
from .utils import canonify, PushBackIterator, in_memory, parse_epoch

"""
Iterators over files on the file system, or in the database, and - building
on that - an iterator over modified files.
"""


def find_stem(path, root, log):
    """
    Find the initial part of path that matches root (and return the length).

    This is needed because we don't know that filenames in the mseedindex table
    are canonified in the same way as here.  So links, etc may make things look
    different even though they are equivalent.
    """
    original_path = path
    while True:
        if canonify(path) == root:
            log.debug('Matched %s as %s' % (root, path))
            return len(path)
        path, _ = split(path)
        if not path:
            raise Exception('Could not find canonical prefix to %s' % original_path)


class DatabasePathIterator(SqliteSupport):
    """
    Ordered iterator over files known in the database.
    We also return last modified as that is used in the indexer.

    The aim here is to be as efficient as possible.  So rather than iterating
    over the filesystem (only) and querying the database for each file, we
    do a single scan of the database and use that to provide a list of known
    files.
    """

    def __init__(self, config):
        super().__init__(config)
        self._data_dir = config.dir(DATADIR)
        self._stem = 0
        self._prev_path = None
        self._cursor = self._db.cursor()
        # prev_path and ordering filemodtime desc guarantee that we get the latest
        # modification time if there's more than one value (not clear if mseedindex
        # updates all rows with the latest value).
        try:
            sql = 'SELECT distinct filemodtime, filename FROM tsindex ORDER BY filename ASC, filemodtime DESC'
            self._log.debug('Execute: %s' % sql)
            self._cursor.execute(sql)
        except OperationalError as e:
            # this is the case when there's no table yet, so no files
            self._cursor = None

    def __iter__(self):
        return self

    def __next__(self):
        try:
            if self._cursor:
                lastmod, path = next(self._cursor)
                if not self._prev_path:
                    self._stem = find_stem(path, self._data_dir, self._log)
                if self._prev_path != path:
                    self._prev_path = path
                return lastmod, join(self._data_dir, path[self._stem + 1:])
            else:
                raise StopIteration()
        except:
            if self._cursor:
                self._cursor.close()
            raise


def RepositoryIterator(root, depth=1):
    """
    Ordered iterator over the filesystem, returning only files from
    the fourth directory level, corresponding to the data files in
    the repository.
    """
    root = canonify(root)
    files = sorted(listdir(root))
    for file in files:
        path = join(root, file)
        if isdir(path) and depth < 4:
            # cannot use 'yield from' as 3to2 doesn't translate it
            for path in RepositoryIterator(path, depth=depth + 1):
                yield path
        elif depth == 4:
            yield path


class ModifiedScanner(SqliteSupport):
    """
    Compare the filesystem and the database (using the iterators above)
    and when there is a discrepancy either remove a database entry or process
    (via subclass) the file.
    """

    def __init__(self, config):
        super().__init__(config)
        self._data_dir = config.dir(DATADIR)
        self._all = config.arg(ALL)
        self._log = config.log
        self._config = config

    def scan_data_dir(self):
        if not exists(self._data_dir):
            makedirs(self._data_dir)
        # pull into memory here to avoid open database when processing
        dbpaths = PushBackIterator(in_memory(DatabasePathIterator(self._config)))
        fspaths = RepositoryIterator(self._data_dir)
        while True:
            closed, dblastmod, dbpath, fspath = False, 0, None, None
            try:
                dblastmod, dbpath = next(dbpaths)
            except StopIteration:
                closed = True
            try:
                fspath = next(fspaths)
            except StopIteration:
                if closed:
                    self.done()
                    return
            # extra entry in file system, so push current database value back,
            # and pretend this entry was in the database, but with a modified date
            # that implies it will be indexed
            if fspath and (not dbpath or fspath < dbpath):
                if not closed:
                    dbpaths.push((dblastmod, dbpath))
                dblastmod, dbpath = '1970-01-01T00:00:00', fspath
            # extra entry in database, needs deleting
            if dbpath and (not fspath or fspath > dbpath):
                self._delete(dbpath)
            # fspath == dbpath so test if need to scan
            else:
                dbepoch = parse_epoch(dblastmod) + 1   # add one because it's rounded down
                if self._all or getmtime(fspath) > dbepoch:
                    self.process(fspath)

    def _delete(self, path):
        self._log.debug('Removing %s from index' % path)
        self.execute('delete from tsindex where filename like ?', (path,))

    def process(self, path):
        raise Exception('Unimplemented')

    def done(self):
        pass


class DirectoryScanner(object):

    def __init__(self, config):
        self._recurse = config.arg(RECURSE)
        self._log = config.log

    def scan_dirs_and_files(self, paths):
        
        for path in paths:
            path = canonify(path)
            if not exists(path):
                raise Exception('Cannot find %s' % path)
            if isfile(path):
                self.process(path) # Calls the process method in rover.ingest        
            else:
                self._scan_dir(path)
        self.done()

    def _scan_dir(self, dir):
        self._log.debug('Scanning directory %s' % dir)
        for file in listdir(dir):
            path = join(dir, file)
            if isfile(path):
                self.process(path) # Calls the process method in rover.ingest
            elif self._recurse:
                self._scan_dir(path)
            else:
                self._log.warn('Ignoring %s in %s (not a file)' % (file, dir))

    def process(self, path):
        #process method could/should be removed. It does nothing and is confusing for the person reading the code . 
        raise Exception('Unimplemented')

    def done(self):
        pass



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import super
from future import standard_library
standard_library.install_aliases()
import datetime
from shutil import copyfile

from rover import __version__
from .args import RETRIEVE, TEMPDIR, AVAILABILITYURL, PREINDEX, LEAP, LEAPEXPIRE, UserFeedback, \
    LEAPFILE, LEAPURL, TEMPEXPIRE, LIST_RETRIEVE, DELETEFILES, POSTSUMMARY, DATASELECTURL, fail_early, HTTPTIMEOUT, \
    HTTPRETRIES, OUTPUT_FORMAT, DATADIR, FORCE_METADATA_RELOAD
from .download import DEFAULT_NAME
from .index import Indexer
from .manager import DownloadManager, ManagerException
from .report import Reporter
from .retrieve_metadata import MetadataRetriever
from .sqlite import SqliteSupport
from .summary import Summarizer
from .utils import clean_old_files, match_prefixes, check_leap, unique_path, \
    safe_unlink, build_file, fix_file_inplace, remove_empty_folders

"""
Commands related to data retrieval:

The `rover retrieve` command - check for remote data that we don't already have, download it and ingest it.
The `rover list-retrieve` command - shows what data would be downloaded by `rover retrieve`.
"""


RETRIEVEWEB = 'rover_retrieve_availability'
RETRIEVECONFIG = 'rover_retrieve_config'
EARLY = datetime.datetime(1900, 1, 1)


class BaseRetriever(SqliteSupport, UserFeedback):
    """
### Retrieve

    rover retrieve file

    rover retrieve [net=N] [sta=S] [loc=L] [cha=C] [start [end]]

    rover retrieve N_S_L_C [start [end]]

Compares ROVER's local index with remotely available data, then downloads and
ingest files missing from the local repository. The URL determining the
availability of remote data can be configured by the availability-url option,
and URL controlling data downloads is configured by the dataselect-url
option.

Use ROVER's list-index function to determine data available on a remote server
which is not in the local repository.

##### Significant Options

@temp-dir
@availability-url
@dataselect-url
@timespan-tol
@pre-index
@ingest
@index
@post-summary
@rover-cmd
@mseedindex-cmd
@data-dir
@download-workers
@download-retries
@http-timeout
@http-retries
@leap-expire
@leap-file
@leap-url
@web
@http-bind-address
@http-port
@email
@email-from
@smtp-address
@smtp-port
@verbosity
@log-dir
@log-verbosity
@temp-expire
@output-format
@asdf-filename
@force-metadata-reload

In addition, options for sub-commands (download, ingest, index) will be used - see help for those
commands for more details.

##### Examples

    rover retrieve N_S_L_C.txt

processes a file containing a request to download, ingest, and index
data missing from ROVER's local repository.

    rover retrieve IU_ANMO_00_BH1 2017-01-01 2017-01-04

processes a command line request to download, ingest, and index
data missing from ROVER's local repository.

"""

    def __init__(self, config):
        UserFeedback.__init__(self, config)
        SqliteSupport.__init__(self, config)
        fail_early(config)
        self._temp_dir = config.dir(TEMPDIR)
        self._availability_url = config.arg(AVAILABILITYURL)
        self._dataselect_url = config.arg(DATASELECTURL)
        self._pre_index = config.arg(PREINDEX)
        self._delete_files = config.arg(DELETEFILES)
        self._post_summary = config.arg(POSTSUMMARY)
        self._download_manager = None   # created in do_run()
        self._reporter = Reporter(config)
        self._config = config
        # leap seconds not used here, but avoids multiple threads all downloading later
        check_leap(config.arg(LEAP), config.arg(LEAPEXPIRE), config.file(LEAPFILE), config.arg(LEAPURL),
                   config.arg(HTTPTIMEOUT), config.arg(HTTPRETRIES), config.log)
        clean_old_files(self._temp_dir, config.arg(TEMPEXPIRE) * 60 * 60 * 24, match_prefixes(RETRIEVEWEB), config.log)

    def do_run(self, args, fetch, command):
        """
        Set-up environment, parse commands, and delegate to sub-methods as appropriate.
        """
        usage = 'Usage: rover %s (file | [net=N] [sta=S] [cha=C] [loc=L] [start [end]] | N_S_L_C [start [end]])' % command
        if not args:
            raise Exception(usage)
        # input is a temp file as we prepend options
        path = unique_path(self._temp_dir, RETRIEVEWEB, args[0])
        try:
            if len(args) == 1:
                copyfile(args[0], path)
            else:
                try:
                    build_file(self._log, path, args)
                except:
                    raise Exception(usage)
            fix_file_inplace(self._log, path, self._temp_dir)
            self._download_manager = DownloadManager(self._config, RETRIEVECONFIG if fetch else None)
            if fetch:
                self._log.default('ROVER version %s - starting retrieve' % __version__)
                self.display_feedback()
            self._query(path, fetch)
            if fetch:
                return self._fetch()
            else:
                return self._display()
        except ManagerException:
            raise
        except Exception as e:
            self._reporter.send_email('ROVER Failure', self._reporter.describe_error(RETRIEVE, e))
            raise
        finally:
            if self._delete_files:
                safe_unlink(path)

    def _query(self, up, fetch):
        """
        Populate the download manager by comparing the data from the
        availability service with the local index.
        """
        if self._pre_index and self._config.arg(OUTPUT_FORMAT).upper() != "ASDF":
            self._log.info('Ensuring index is current before retrieval')
            Indexer(self._config).run([])
        self._download_manager.add(DEFAULT_NAME, up, fetch,
                                   self._availability_url, self._dataselect_url, self._source_callback)

    def _fetch(self):
        """
        Fetch data from the download manager.
        """
        try:
            n_downloads = self._download_manager.download()
        finally:
            if self._post_summary:
                Summarizer(self._config).run([])
                self._complete_asdf_retrieve()
        return n_downloads

    def _complete_asdf_retrieve(self):
        if self._config.arg(OUTPUT_FORMAT).upper() == "ASDF":
            # remove empty mseed directories from data directory
            remove_empty_folders(self._config.arg(DATADIR), self._log)
            # load metadata into asdf dataset
            MetadataRetriever(self._config).run([])

    def _display(self):
        """
        Display data from the download manager.
        """
        return self._download_manager.display()

    def _source_callback(self, source):
        subject, msg = self._reporter.describe_retrieve(source)
        self._reporter.send_email(subject, msg)


class Retriever(BaseRetriever):

    __doc__ = BaseRetriever.__doc__

    def __init__(self, config):
        super().__init__(config)

    def run(self, args):
        return self.do_run(args, True, RETRIEVE)


class ListRetriever(BaseRetriever):
    """
### List Retrieve

    rover list-retrieve file

    rover list-retrieve N_S_L_C [start [end]]

Compares the local index with the requested data remotely available, then
displays the difference. Note that the summary is printed to stdout, while
logging is to stderr.

##### Significant Options

@availability-url
@timespan-tol
@data-dir
@verbosity
@log-dir
@log-verbosity

##### Examples

    rover list-retrieve N_S_L_C.txt

will display the data missing form the repository to match what is available for the stations in the given file.

    rover list-retrieve IU.ANMO.00.BH1 2017-01-01 2017-01-04

will display the data missing from the repository to match what is available for IU.ANMO.00.BH1.

"""

    def __init__(self, config):
        super().__init__(config)

    def run(self, args):
        return self.do_run(args, False, LIST_RETRIEVE)



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from builtins import super
from builtins import range
from builtins import int
from builtins import map
from future import standard_library
standard_library.install_aliases()
from shutil import copyfile
from sqlite3 import OperationalError

from .request import RequestComparison
from .manager import DownloadManager
from .args import SUBSCRIBE, LIST_SUBSCRIBE, UNSUBSCRIBE, SUBSCRIPTIONSDIR, AVAILABILITYURL, DATASELECTURL, DEV, \
    FORCEREQUEST, mm, TRIGGER, VERBOSITY, NO, DELETEFILES, TEMPDIR
from .sqlite import SqliteSupport, NoResult
from .utils import unique_path, build_file, format_day_epoch, safe_unlink, format_time_epoch, log_file_contents, \
    fix_file_inplace

"""
Commands related to subscription:

The 'rover subscribe' command - add a subscription
The 'rover unsubscribe' command - remove a subscription
The 'rover list-subscriptions' command - display information about subscriptions
"""


SUBSCRIBEFILE = 'rover_subscribe'


class Subscriber(SqliteSupport):
    """
### Subscribe

    rover subscribe file

    rover subscribe [net=N] [sta=S] [loc=L] [cha=C] [start [end]]

    rover subscribe N_S_L_C [start [end]]

Subscribe generates a background service, daemon, that regularly compares data
available at the configured server with the local repository. If there is a
discrepancy, available data is downloaded, ingested and indexed. ROVER subscribe
is similar to `rover retrieve` but uses a daemon to regularly update a local
repository.

##### Significant Options

@availability-url
@dataselect-url
@force-request
@verbosity
@log-dir
@log-verbosity
@temp-dir

Most of the download process is controlled by the options provided when starting the service (see
`rover start`).

##### Examples

    rover subscribe N_S_L_C.txt

will instruct the daemon to regularly download, ingest, and index any data missing from the repository for
NSLCs / timespans in the given file.

    rover subscribe IU_ANMO_00_BH1 2017-01-01 2017-01-04

will instruct the daemon to regularly download, ingest and index and data for IU.ANMO.00.BH1 between the given
dates that are missing from the repository.

    """

    def __init__(self, config):
        super().__init__(config)
        self._force_request = config.arg(FORCEREQUEST)
        self._subscriptions_dir = config.dir(SUBSCRIPTIONSDIR)
        self._availability_url = config.arg(AVAILABILITYURL)
        self._dataselect_url = config.arg(DATASELECTURL)
        self._temp_dir = config.dir(TEMPDIR)
        self._create_table()

    def _create_table(self):
        self.execute('''CREATE TABLE IF NOT EXISTS rover_subscriptions (
                           id integer primary key autoincrement,
                           file text unique,
                           availability_url text not null,
                           dataselect_url text not null,
                           creation_epoch int default (cast(strftime('%s', 'now') as int)),
                           last_check_epoch int default NULL,
                           last_error_count int default 0,
                           consistent int default 0
        )''')

    def _check_all_for_overlap(self, path1):
        rows = self.fetchall('''SELECT file FROM rover_subscriptions''')
        for row in rows:
            try:
                RequestComparison(path1, row[0]).assert_no_overlap()
            except Exception as e:
                self._log.error('To avoid duplicating data, subscriptions for the same N_S_L_C and '+
                                'time range are not allowed.')
                self._log.error('Your latest subscription appears to overlap an existing subscription ' +
                                '(see error below).')
                self._log.error('The first 10 lines of your new subscription are:')
                log_file_contents(path1, self._log, 10)
                self._log.error('The first 10 lines of the existing subscription are:')
                log_file_contents(row[0], self._log, 10)
                raise

    def run(self, args):
        # input is a temp file as we prepend options
        try:
            path = unique_path(self._subscriptions_dir, SUBSCRIBEFILE, args[0])
            if len(args) == 1:
                copyfile(args[0], path)
            else:
                build_file(self._log, path, args)
            fix_file_inplace(self._log, path, self._temp_dir)
        except:
            raise Exception('Usage: rover %s (file | [net=N] [sta=S] [cha=C] [loc=L] [start [end]] | N_S_L_C [start [end]])' % SUBSCRIBE)
        if self._force_request:
            self._log.warn('Not checking for overlaps (%s) - may result in duplicate data in the repository' % (mm(FORCEREQUEST)))
        else:
            self._check_all_for_overlap(path)
        self.execute('''INSERT INTO rover_subscriptions (file, availability_url, dataselect_url) VALUES (?, ?, ?)''',
                     (path, self._availability_url, self._dataselect_url))
        self._log.default('Subscribed')


def parse_integers(args):
    # return a list rather than a generator so we fail fast
    ids = []
    for arg in args:
        if ':' in arg:
            try:
                id1, id2 = map(int, arg.split(':'))
            except:
                raise Exception('Cannot parse %s as a pair of IDs' % arg)
        else:
            try:
                id = int(arg)
                id1, id2 = id, id
            except:
                raise Exception('Cannot parse %s as an IDs' % arg)
        ids.append((id1, id2))
    return ids


class SubscriptionLister(SqliteSupport):
    """
### List Subscribe

    rover list-subscribe

    rover list-subscribe N[:M]+

The first form, with no arguments, lists the current subscriptions (these can be removed with `rover unsubscribe`).

The second form, similar to `rover list-retrieve`, shows the data that will be downloaded for that subscription.
The arguments can be single numbers (identifying the subscriptions, as displayed by `rover list-subscrive`), or
ranges (N:M).

The data to be downloaded is not exact, because the daemon may already be downloading data, or because when
the subscription is processed (in the future) the available data have changed.

##### Significant Options

@data-dir
@verbosity
@log-dir
@log-verbosity

##### Examples

    rover list-subscribe

will list the subscriptions and the IDs.

    rover list-subscribe ID

will show what data would be downloaded by the daemon if the subscription were processed immediately.

    """

    def __init__(self, config):
        super().__init__(config)
        self._config = config
        self._dev = config.arg(DEV)

    def run(self, args):
        try:
            ids = parse_integers(args)
        except:
            raise Exception('Usage: rover %s [id | id1:id2] ...' % LIST_SUBSCRIBE)
        if not ids:
            self._list_subscriptions()
        else:
            self._list_downloads(ids)

    def _list_downloads(self, ids):
        download_manager = DownloadManager(self._config)
        for (id1, id2) in ids:
            for id in range(id1, id2+1):
                try:
                    path, availability_url, dataselect_url = self.fetchone(
                        '''SELECT file, availability_url, dataselect_url FROM rover_subscriptions WHERE id = ?''', (id,),
                    quiet=True)
                    self.add = download_manager.add(id, path, False, availability_url, dataselect_url, None)
                except (NoResult, OperationalError):
                    self._log.warn('No subscription %d' % id)
        download_manager.display()

    def _list_subscriptions(self):
        print()
        count = [0]

        def callback(row):
            id, file, availability_url, dataselect_url, creation_epoch, last_check_epoch = row
            date = format_day_epoch(creation_epoch)
            try:
                check = format_time_epoch(last_check_epoch)
            except TypeError:
                check = 'never'
            print('  %d created %s  checked %s' % (id, date, check))
            print('    %s' % file)
            print('    %s' % availability_url)
            print('    %s' % dataselect_url)
            print()
            count[0] += 1

        try:

            self.foreachrow('''SELECT id, file, availability_url, dataselect_url, creation_epoch, last_check_epoch
                                 FROM rover_subscriptions ORDER BY creation_epoch
                            ''', tuple(), callback, quiet=True)
        except OperationalError:
            # no table
            if self._dev: raise

        print("  %d subscription%s" % (count[0], '' if count[0] == 1 else 's'))
        print()


class Unsubscriber(SqliteSupport):
    """
### Unsubscribe

    rover unsubscribe N[:M]+

Deletes one or more subscriptions identified by their indices.
ROVER list-subscribe displays subscription indices. Unsubscribe accepts
integers or ranges of integers (N:M) as arguments. Data associated with a
subscription(s) are not deleted.

To avoid conflicts with subscriptions that are currently being processed,
`rover stop` must stop the daemon before using the unsubscribe command.

##### Significant Options

@data-dir
@verbosity
@log-dir
@log-verbosity

##### Examples

    rover unsubscribe 1:3

will delete subscriptions 1, 2 and 3.

    """

    def __init__(self, config):
        super().__init__(config)

    def run(self, args):
        if not args:
            raise Exception('Usage: rover %s (id|id1:id2)+' % UNSUBSCRIBE)
        for id1, id2 in parse_integers(args):

            def callback(row):
                file = row[0]
                self._log.debug('Deleting %s' % file)
                safe_unlink(file)

            self.foreachrow('''SELECT file FROM rover_subscriptions WHERE id >= ? AND id <= ?''', (id1, id2), callback)
            self.execute('''DELETE FROM rover_subscriptions WHERE id >= ? AND id <= ?''', (id1, id2))
            self._log.default('Cleared subscriptions between %d and %d' % (id1, id2))


class Trigger(SqliteSupport):
    """
### Trigger

    rover trigger N[:M]+

Ask the daemon to immediately re-process a subscription(s) based on its index.
List-subscribe displays subscriptions indices. Trigger accepts integers or
ranges of integers (N:M) as arguments.

#### Significant Options

@data-dir
@verbosity
@log-dir
@log-verbosity

##### Examples

    rover trigger 2

will ask the daemon to re-process subscription 2.

    """

    def __init__(self, config):
        super().__init__(config)

    def run(self, args):
        if not args:
            raise Exception('Usage: rover %s (id|id1:id2)+' % TRIGGER)
        for id1, id2 in parse_integers(args):
            self.execute('''UPDATE rover_subscriptions SET last_check_epoch = NULL WHERE id >= ? AND id <= ?''',
                         (id1, id2))
            self._log.default('Cleared last check date for subscriptions between %d and %d' % (id1, id2))



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import super
from future import standard_library
standard_library.install_aliases()
from os import getpid
from sqlite3 import OperationalError, IntegrityError
from time import sleep
from .utils import format_epoch, process_exists
from .sqlite import SqliteSupport


"""
Locking of named resources via the database.
"""


# name used for locking the mseed data file
MSEED = "mseed"
# name used for locking the asdf data file
ASDF = "asdf"


class DatabaseBasedLockFactory(SqliteSupport):
    """
    Support locking against some string (eg name of file) via the database.
    We're trying to avoid file locking because of NFS and cross-platform issues, so use this instead.
    A new instance of the factory must be generated for each kind of resource locked.  Multiple instances
    can exist for the same kind of resource (in different processes) - that's the whole point.
    """

    def __init__(self, config, name):
        super().__init__(config)
        self._table_name = 'rover_lock_%s' % name
        self._config = config
        self._create_lock_table()

    def _create_lock_table(self):
        self.execute('''CREATE TABLE IF NOT EXISTS %s (
                           id integer primary key autoincrement,
                           pid integer unique,
                           key text unique,
                           creation_epoch int default (cast(strftime('%%s', 'now') as int))
        )''' % self._table_name)

    def lock(self, key, pid=None):
        return LockContext(self._config, self._table_name, key, pid)


class LockContext(SqliteSupport):
    """
    Both context and acquire/release syntax are supported here.
    """

    def __init__(self, config, table, key, pid):
        super().__init__(config)
        self._table = table
        self._key = key
        self._pid = pid

    def __enter__(self):
        self.acquire()

    def acquire(self):
        clean = False
        while True:
            try:
                if clean:
                    if not self._clean():
                        self._log.debug('Sleeping on lock %s with %s' % (self._table, self._key))
                        sleep(1)
                # very careful with transactions here - want entire process to be in a single transaction
                with self._db:  # commits or rolls back
                    c = self._db.cursor()
                    c.execute('BEGIN')
                    if not c.execute('SELECT count(*) FROM %s WHERE key = ?' % self._table, (self._key,)).fetchone()[0]:
                        self._log.debug('Acquiring lock on %s with %s for PID %d' % (self._table, self._key, getpid()))
                        c.execute('INSERT INTO %s (pid, key) VALUES (?, ?)' % self._table, (self._pid, self._key))
                        return
            except IntegrityError as e:
                self._log.debug('Acquiring lock: %s' % e)
                sleep(1)
                pass  # PID existed and needs to be cleaned out
            except OperationalError as e:
                self._log.debug('Acquiring lock: %s' % e)
                sleep(1)
                pass  # database was locked
            clean = True

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.release()
        return False

    def set_pid(self, pid):
        self._log.debug('Setting PID on %s for %s to %d' % (self._table, self._key, pid))
        with self._db:
            c = self._db.cursor()
            c.execute('BEGIN')
            c.execute('UPDATE %s SET pid=? WHERE key=?' % self._table, (pid, self._key))

    def release(self):
        self._log.debug('Releasing lock on %s with %s' % (self._table, self._key))
        with self._db:
            c = self._db.cursor()
            c.execute('BEGIN')
            c.execute('DELETE FROM %s WHERE key = ?' % self._table, (self._key,))

    def _clean(self):
        cleaned = [False]

        def callback(row):
            pid, key, epoch = row
            if pid is not None and not process_exists(pid):
                # no longer a warning.  seems to occur when two transactions both
                # delete the same entry (one from the worker existing and one from a worker
                # waiting).  waiting starts the transaction before exiting, but completes
                # after, afaict (so the exiting has disappeared and the PID test succeeds
                # for the waiting),
                self._log.debug('Cleaning out old entry for PID %d on lock %s with %s (created %s)' % (
                    pid, self._table, key, format_epoch(epoch)))
                with self._db:
                    c = self._db.cursor()
                    c.execute('BEGIN')
                    c.execute('DELETE FROM %s WHERE key = ?' % self._table, (self._key,))
                    cleaned[0] = True

        self.foreachrow('SELECT pid, key, creation_epoch FROM %s' % self._table, tuple(), callback)
        return cleaned[0]



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from builtins import super
from future import standard_library
standard_library.install_aliases()
import sys
from re import match, sub
from sqlite3 import OperationalError

from .config import timeseries_db
from .index import START, END
from .utils import STATION, NETWORK, CHANNEL, LOCATION, format_epoch, tidy_timestamp
from .args import SUMMARY
from .sqlite import SqliteSupport


"""
Commands related to the summary:

The 'rover summary' command - creates a summary of the tsindex table.
The 'rover list-summary' command - displays entries from the summary table.
"""


class Summarizer(SqliteSupport):
    """
### Summary

    rover summary

Creates a summary of the index stored in a ROVER repository. This lists the
overall span of data for each Net_Sta_Loc_Chan and can be queried using
`rover list-summary`.

##### Significant Options

@data-dir
@verbosity
@log-dir
@log-verbosity

##### Examples

    rover summary

will create the summary of a local ROVER repository.


    """

    def __init__(self, config):
        super().__init__(config)

    def run(self, args):
        if len(args):
            raise Exception('Usage: rover %s' % SUMMARY)
        self._log.info('Generating summary table')
        self.execute('DROP TABLE IF EXISTS tsindex_summary')
        try:
            self.execute('''CREATE TABLE tsindex_summary AS
                              SELECT network, station, location, channel,
                                     min(starttime) AS earliest, max(endtime) AS latest,
                                     datetime('now') AS updt
                              FROM tsindex
                              GROUP BY 1,2,3,4''')
        except OperationalError:
            self._log.default('No index found')


class SummaryLister(SqliteSupport):
    """
### List Summary

    rover list-summary [net=...|sta=...|loc=...|cha=..]* [start=...] [end=...]

    rover list-summary [N_S_L_C_Q]* [start=...] [end=...]

List a summary of entries for a ROVER repository, defined by the data-dir
configuration option, that match given constraints. List summary is faster
than `rover list-index` but gives less detail. For more information,
run "rover list-index" with no arguments.

##### Significant Options

@data-dir
@verbosity
@log-dir
@log-verbosity

#### Examples

    rover list-summary net=* start=2001-01-01

list all entries in the summary after 2001-01-01.

"""

    # this is suspiciously close to list-index, but the differences are significant and
    # to make a common base class would likely have made things more opaque.

    def __init__(self, config):
        SqliteSupport.__init__(self, config)
        self._timeseries_db = timeseries_db(config)
        self._multiple_constraints = {STATION: [],
                                      NETWORK: [],
                                      CHANNEL: [],
                                      LOCATION: []}
        self._single_constraints = {START: None,
                                    END: None}

    def run(self, args, stdout=None):
        if not stdout:
            stdout = sys.stdout
        self._check_database()
        self._parse_args(args)
        sql, params = self._build_query()
        self._rows(sql, params, stdout=stdout)

    def _check_database(self):
        try:
            self.execute('SELECT count(*) FROM tsindex_summary')
        except OperationalError:
            raise Exception('''Cannot access the summary table in the database (%s).  Bad configuration or summary not generated?''' % self._timeseries_db)

    def _parse_args(self, args):
        for arg in args:
            parts = arg.split('=')
            if len(parts) == 1:
                self._set_snclq(arg)
            elif len(parts) != 2:
                raise Exception('Cannot parse "%s" (not of form name=value)' % arg)
            else:
                name, value = parts
                if name in (START, END):
                    self._set_time_limit(name, value)
                else:
                    self._set_name_value(name, value)

    def _set_snclq(self, snclq):
        components = snclq.split('_')
        if len(components) < 2 or len(components) > 4:
            raise Exception('Cannot parse %s (expect 2 to 4 values separated by "_" for N_S_L_C)' % snclq)
        self._set_name_value(NETWORK, components[0])
        self._set_name_value(STATION, components[1])
        if len(components) > 2: self._set_name_value(LOCATION, components[2])
        if len(components) > 3: self._set_name_value(CHANNEL, components[3])

    def _set_time_limit(self, name, value):
        if self._single_constraints[name]:
            raise Exception('Multiple values for %s' % name)
        value = tidy_timestamp(self._log, value)
        self._single_constraints[name] = value
        if self._single_constraints[START] and self._single_constraints[END] \
                and self._single_constraints[START] > self._single_constraints[END]:
            raise Exception('start time (%s) must be after end time (%s)'
                            % (self._single_constraints[START], self._single_constraints[END]))

    def _set_name_value(self, name, value):
        if not match('^[\w\*\?]*$', value):
            raise Exception('Illegal characters in "%s"' % value)
        found = None
        for key in self._multiple_constraints.keys():
            if key.startswith(name):
                if found:
                    raise Exception('Ambiguous option: %s' % name)
                else:
                    found = key
        self._multiple_constraints[found].append(value)

    def _build_query(self):
        sql, params = 'SELECT network, station, location, channel, earliest, latest FROM tsindex_summary ', []
        constrained = False

        def conjunction(sql, constrained):
            if constrained:
                sql += 'and '
            else:
                sql += 'where '
                constrained = True
            return sql, constrained

        for name in self._multiple_constraints:
            repeated = False
            for value in self._multiple_constraints[name]:
                if not repeated:
                    sql, constrained = conjunction(sql, constrained)
                    sql += '( '
                    repeated = True
                else:
                    sql += 'or '
                sql += '%s like ? ' % name
                params.append(self._wildchars(value))
            if repeated:
                sql += ') '
        if self._single_constraints[START]:
            sql, constrained = conjunction(sql, constrained)
            sql += 'latest > ?'
            params.append(self._single_constraints[START])
        if self._single_constraints[END]:
            sql, constrained = conjunction(sql, constrained)
            sql += 'earliest < ?'
            params.append(self._single_constraints[END])
        return sql, tuple(params)

    @staticmethod
    def _wildchars(value):
        return sub(r'\?', '_', sub(r'\*', '%', value))

    def _rows(self, sql, params, stdout):
        self._log.debug('%s %s' % (sql, params))

        def callback(row):
            n, s, l, c, b, e = row
            print('  %s_%s_%s_%s %s %s' % (n, s, l, c, b, e))

        print()
        self.foreachrow(sql, params, callback)
        print()



# For reference: https://github.com/pypa/sampleproject/blob/master/setup.py

from setuptools import setup, find_packages
from io import open
import os

module_name = 'rover'

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, "README.md"), encoding='utf-8') as fh:
    long_description = fh.read()

with open(os.path.join(here, module_name, 'VERSION')) as vf:
    version = vf.read().strip()

setup(
    name=module_name,
    version=version,
    author="EarthScope",
    author_email="software@earthscope.org",
    description="Tool for the robust retrieval of seismic data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://earthscope.github.io/rover",
    packages=find_packages(),
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, <4',
    classifiers=(
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: Unix",
        "Operating System :: MacOS",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ),
    install_requires=["requests", "future"],
    extras_require={
        'dev': ["nose", "robotframework"],
        'mseedindex': ["mseedindex"],
    },
    entry_points={
        'console_scripts': [
            '%s = %s:main' % (module_name, module_name),
        ],
    },
    package_data={
        module_name: ['VERSION']
    }
)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import super
from builtins import next
from future import standard_library
standard_library.install_aliases()
from os import getpid, kill

from .args import RETRIEVE, DAEMON, START, STOP, UNSUBSCRIBE
from .sqlite import SqliteSupport, NoResult
from .utils import process_exists


"""
Process management (mainly stopping duplicate data by refusing to run retrieve and the daemon in parallel).
"""


class ProcessManager(SqliteSupport):

    def __init__(self, config):
        super().__init__(config)
        self._command = config.command
        self._create_processes_table()

    def _create_processes_table(self):
        self.execute('''CREATE TABLE IF NOT EXISTS rover_processes (
                           id integer primary key autoincrement,
                           pid integer unique,
                           command text not null,
                           creation_epoch int default (cast(strftime('%s', 'now') AS int))
        )''')

    def __enter__(self):
        """
        Check whether the command should run
        """
        # retrieve and the daemon cannot run in parallel because they may try to
        # update the same data.
        # unsubscribe requires that teh daemon be stopped since it may currently be
        # being processed.
        # any other command is either harmless (list...) or likely done by an expert user
        self._log.debug('Process check %s' % self._command)
        if self._command in (RETRIEVE, UNSUBSCRIBE):
            self._check_command(self._command)
        # we don't need to check start, because the daemon it spawns will be checked, but
        # it's better to have the error early and visible to the user
        elif self._command in (DAEMON, START):
            self._check_daemon(self._command == DAEMON)

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._command in (RETRIEVE, DAEMON):
            self._clean_entry(self._command)
        return False  # let any exception propagate

    def _current_command_inside_transaction(self):
        pid, command = None, None
        rows = self._db.execute('SELECT pid, command FROM rover_processes', tuple())
        try:
            pid, candidate = next(rows)
            if process_exists(pid):
                command = candidate
                self._log.debug('Current process is %s/%d' % (command, pid))
            else:
                self._log.debug('Removing dead process %s/%d' % (candidate, pid))
                self._db.execute('DELETE FROM rover_processes WHERE pid = ?', (pid,))
        except StopIteration:
            pass  # table is empty
        try:
            next(rows)
            raise Exception('Multiple entries in rover_processes')
        except StopIteration:
            pass  # no more entries
        return pid, command

    def _record_process_inside_transaction(self, command):
        self._log.debug('Record new process %s/%d' % (command, getpid()))
        self._db.execute('INSERT INTO rover_processes (command, pid) VALUES (?, ?)', (command, getpid()))

    def _check_command(self, name):
        error = None
        with self._db:  # single transaction
            self._db.cursor().execute('BEGIN')
            pid, command = self._current_command_inside_transaction()
            if command == DAEMON:
                error = Exception(('You cannot use rover %s while the %s is running (PID %d). ' +
                                   'Use "rover %s" to stop the daemon.') %  (name, DAEMON, pid, STOP))
            elif command == RETRIEVE:
                error = Exception('rover %s is already running (PID %d).' % (RETRIEVE, pid))
            else:
                self._record_process_inside_transaction(RETRIEVE)
        # transaction closed
        if error:
            raise error

    def _check_daemon(self, record):
        error = None
        with self._db:  # single transaction
            self._db.cursor().execute('BEGIN')
            pid, command = self._current_command_inside_transaction()
            if command == RETRIEVE:
                error = Exception('You cannot use the %s while rover %s is running (PID %d). ' %
                                  (DAEMON, RETRIEVE, pid))
            elif command == DAEMON:
                error = Exception('The %s is already running (PID %d).' % (DAEMON, pid))
            elif record:
                self._record_process_inside_transaction(DAEMON)
        # transaction closed
        if error:
            raise error

    def _delete_command_inside_transaction(self, command):
        self._db.execute('DELETE FROM rover_processes WHERE command LIKE ?', (command,))

    def _clean_entry(self, command):
        with self._db:
            self._db.cursor().execute('BEGIN')
            self._delete_command_inside_transaction(command)

    def current_command(self):
        with self._db:
            self._db.cursor().execute('BEGIN')
            pid, command = self._current_command_inside_transaction()
            return pid, command

    def _pid_inside_transaction(self, command):
        cmd, params = 'SELECT pid FROM rover_processes WHERE command LIKE ?', (command,)
        try:
            pid = next(self._db.execute(cmd, params))[0]
            if process_exists(pid):
                return pid
            else:
                self._delete_command_inside_transaction(command)
        except Exception:
            pass
        raise NoResult(cmd, params)

    def kill_daemon(self):
        """
        Kill the daemon, if it exists.
        """
        try:
            with self._db:
                self._db.cursor().execute('BEGIN')
                pid = self._pid_inside_transaction(DAEMON)
                self._log.default('Killing %s (pid %d)' % (DAEMON, pid))
                kill(pid, 9)
                self._delete_command_inside_transaction(DAEMON)
        except NoResult:
            raise Exception('The %s is not running' % DAEMON)

    def daemon_status(self):
        with self._db:
            try:
                self._db.cursor().execute('BEGIN')
                pid = self._pid_inside_transaction(DAEMON)
                return 'The %s is running (process %d)' % (DAEMON, pid)
            except NoResult:
                return 'The %s is not running' % DAEMON




from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from builtins import open
from builtins import super
from builtins import str
from builtins import object
from future import standard_library
standard_library.install_aliases()
import sys
from argparse import ArgumentParser, Action, RawDescriptionHelpFormatter, SUPPRESS
from os.path import exists, join
from re import sub
from smtplib import SMTP_PORT
from textwrap import dedent

from rover import __version__
from .utils import create_parents, canonify, check_cmd, dictionary_text_list

"""
Command line / file configuration parameters.
"""

ERROR_CODE = 1
ABORT_CODE = 2

# commands
DAEMON = 'daemon'
DOWNLOAD = 'download'
HELP_CMD = 'help'
INDEX = 'index'
INGEST = 'ingest'
LIST_INDEX = 'list-index'
LIST_RETRIEVE = 'list-retrieve'
LIST_SUBSCRIBE = 'list-subscribe'
LIST_SUMMARY = 'list-summary'
RETRIEVE = 'retrieve'
RETRIEVE_METADATA = 'retrieve-metadata'
START = 'start'
STOP = 'stop'
STATUS = 'status'
SUMMARY = 'summary'
SUBSCRIBE = 'subscribe'
TRIGGER = 'trigger'
UNSUBSCRIBE = 'unsubscribe'
WEB = 'web'
INIT = 'init'
INIT_REPO = 'init-repo'
INIT_REPOSITORY = 'init-repository'


# flag negation
NO = '--no-'


# parameters
ALL = 'all'
ARGS = 'args'
ASDF_FILENAME = 'asdf-filename'
AVAILABILITYURL = 'availability-url'
COMMAND = 'command'
DATADIR = 'data-dir'
DATASELECTURL = 'dataselect-url'
DELETEFILES = 'delete-files'
DOWNLOADRETRIES = 'download-retries'
DOWNLOADWORKERS = 'download-workers'
DEV = 'dev'
EMAIL = 'email'
EMAILFROM = 'email-from'
F, FILE = 'f', 'file'
FORCECMD = 'force-cmd'
FORCEFAILURES = 'force-failures'
FORCE_METADATA_RELOAD = 'force-metadata-reload'
FORCEREQUEST = 'force-request'
H, FULLHELP = 'H', 'full-help'
_h, _help = 'h', 'help'
FULLCONFIG = 'full-config'
HTTPBINDADDRESS = 'http-bind-address'
HTTPPORT = 'http-port'
HTTPRETRIES = 'http-retries'
HTTPTIMEOUT = 'http-timeout'
LEAP = 'leap'
LEAPEXPIRE = 'leap-expire'
LEAPFILE = 'leap-file'
LEAPURL = 'leap-url'
LOGDIR = 'log-dir'
LOGVERBOSITY = 'log-verbosity'
LOGSIZE = 'log-size'
LOGUNIQUE = 'log-unique'
LOGUNIQUEEXPIRE = 'log-unique-expire'
LOGCOUNT = 'log-count'
MDFORMAT = 'md-format'
MSEEDINDEXCMD = 'mseedindex-cmd'
MSEEDINDEXWORKERS = 'mseedindex-workers'
OUTPUT_FORMAT = 'output-format'
POSTSUMMARY = 'post-summary'
PREINDEX = 'pre-index'
RECHECKPERIOD = 'recheck-period'
RECURSE = "recurse"
ROVERCMD = 'rover-cmd'
SMTPADDRESS = 'smtp-address'
SMTPPORT = 'smtp-port'
SORTINPYTHON = 'sort-in-python'
STATIONURL = 'station-url'
SUBSCRIPTIONSDIR = 'subscriptions-dir'
TEMPDIR = 'temp-dir'
TEMPEXPIRE = 'temp-expire'
TIMESPANINC = 'timespan-inc'
TIMESPANTOL = 'timespan-tol'
LITTLE_V, VERBOSITY = 'v', 'verbosity'
BIG_V, VERSION = 'V', 'version'

LITTLE_HELP = (DOWNLOADRETRIES, LOGDIR, VERBOSITY, EMAIL,
               VERSION, HELP_CMD, FULLHELP, FILE, AVAILABILITYURL, DATASELECTURL)
LITTLE_CONFIG = (DATADIR, DOWNLOADRETRIES, DOWNLOADWORKERS, OUTPUT_FORMAT,
                 ASDF_FILENAME, STATIONURL, AVAILABILITYURL, DATASELECTURL,
                 TEMPDIR, LOGDIR, LOGVERBOSITY, VERBOSITY, WEB, HTTPPORT,
                 EMAIL, SMTPADDRESS, SMTPPORT)
DYNAMIC_ARGS = (VERSION, HELP_CMD, FULLHELP)

# default values (for non-boolean parameters)
DEFAULT_ASDF_FILENAME = 'asdf.h5'
DEFAULT_AVAILABILITYURL = 'http://service.iris.edu/fdsnws/availability/1/query'
DEFAULT_DATADIR = 'data'
DEFAULT_DATASELECTURL = 'http://service.iris.edu/fdsnws/dataselect/1/query'
DEFAULT_DOWNLOADRETRIES = 3
DEFAULT_DOWNLOADWORKERS = 5
DEFAULT_EMAILFROM = 'noreply@rover'
DEFAULT_FILE = join('rover.config')
DEFAULT_FORCEFAILURES = 0
DEFAULT_HTTPBINDADDRESS = '127.0.0.1'
DEFAULT_HTTPPORT = 8000
DEFAULT_HTTPRETRIES = 3
DEFAULT_HTTPTIMEOUT = 60
DEFAULT_LEAPEXPIRE = 30
DEFAULT_LEAPFILE = 'leap-seconds.list'
DEFAULT_LEAPURL = 'https://www.ietf.org/timezones/data/leap-seconds.list'
DEFAULT_LOGDIR = 'logs'
DEFAULT_LOGVERBOSITY = 4
DEFAULT_LOGSIZE = '10M'
DEFAULT_LOGCOUNT = 10
DEFAULT_LOGUNIQUE_EXPIRE = 7
DEFAULT_MSEEDINDEXCMD = 'mseedindex -sqlitebusyto 60000'
DEFAULT_MSEEDINDEXWORKERS = 10
DEFAULT_OUTPUT_FORMAT = 'mseed'
DEFAULT_RECHECKPERIOD = 12
DEFAULT_ROVERCMD = 'rover'
DEFAULT_SMTPADDRESS = 'localhost'
DEFAULT_STATIONURL = 'http://service.iris.edu/fdsnws/station/1/query'
DEFAULT_SUBSCRIPTIONSDIR = 'subscriptions'
DEFAULT_TEMPDIR = 'tmp'
DEFAULT_TEMPEXPIRE = 1
DEFAULT_TIMESPANINC = 0.5
DEFAULT_TIMESPANTOL = 0.5
DEFAULT_VERBOSITY = 4


DIRVAR = 'DIR'
FILEVAR = 'FILE'
SAMPLESVAR = 'SAMPLES'
NVAR = 'N'
CMDVAR = 'CMD'
DAYSVAR = 'DAYS'
SECSVAR = 'SECS'
HOURSVAR = 'HOURS'
PERCENTVAR = 'PERCENT'
ADDRESSVAR = 'ADDRESS'
URLVAR = 'URL'
SIZE = 'SIZE'

def parse_bool(value):
    """
    Treat caseless true, yes and on and true, anything else as false.
    """
    if not value:
        value = 'False'
    value = value.lower()
    return value in ('true', 'yes', 'on')


class StoreBoolAction(Action):
    """
    We need a special action for booleans because we must covertly
    parse them as '--foo True' even though the user only types '--foo'.
    """

    def __init__(self,
                 option_strings,
                 dest,
                 default=None,
                 choices=None,
                 required=False,
                 help=None,
                 metavar=None):
        super().__init__(
            option_strings=option_strings,
            dest=dest,
            nargs=1,
            const=None,
            default=default,
            type=parse_bool,
            choices=choices,
            required=required,
            help=help,
            metavar=metavar)

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, values[0])


class FullHelpAction(Action):
    """
    Associate -H or --full-help with all help details.
    """

    def __init__(self,
                 option_strings,
                 dest=SUPPRESS,
                 default=SUPPRESS,
                 help=None):
        super(FullHelpAction, self).__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help)

    def __call__(self, parser, namespace, values, option_string=None):
        parser.print_big_help()
        parser.exit()



def m(string): return '-' + string


def mm(string): return '--' + string


def unbar(string): return sub('_', '-', string)


def cmd_or_alias(cmd):
    """
    We use the 'type' of the command argument to handle aliases.  In this
    way we avoid having to have multiple checks elsewhere in the startup
    logic.
    """
    if cmd in (INIT, INIT_REPO):
        cmd = INIT_REPOSITORY
    return cmd


class Arguments(ArgumentParser):
    """
    Extend the standard arg parsing to:
    * scan initial args to find if config file location specifed
    * if config file is missing, generate defaults
    * read config file before command line args

    To do this correctly for boolean flags (which behave differently on
    the command line and in the config file) we also have to pre-process those.

    The aim is to:
    * have (almost) all config duplcaited, both in config and command line
    * have default config be self-documenting and discoverable
    """

    def __init__(self):
        super().__init__(fromfile_prefix_chars='@', prog='ROVER',
                         formatter_class=RawDescriptionHelpFormatter,
                         description='ROVER: Retrieval of Various Experiment data Robustly',
                         add_help=False,
                         epilog=dedent('''
                         Flags can be negated (eg --no-web).
                         Defaults are read from the configuration file (%s).
                         Type "rover help" for more information on available commands.''' % DEFAULT_FILE))
        self.register('action', 'store_bool', StoreBoolAction)
        # operation details
        operation_group = self.add_argument_group('general operation arguments')
        operation_group.add_argument(m(_h), mm(_help), action='store_true', help='show this help message and exit')
        operation_group.add_argument(m(H), mm(FULLHELP), action=FullHelpAction, help='show full help details')
        operation_group.add_argument(m(BIG_V), mm(VERSION), action='version', version='ROVER %s' % __version__)
        operation_group.add_argument(mm(FULLCONFIG), default=False, action='store_bool', help='initialize with full configuration file', metavar='')
        operation_group.add_argument(m(F), mm(FILE), default=DEFAULT_FILE, help='specify configuration file')
        # metavar must be empty string to hide value since user options
        # are flags that are automatically given values below.
        operation_group.add_argument(mm(DEV), default=False, action='store_bool', help='development mode (show exceptions)?', metavar='')
        operation_group.add_argument(mm(DELETEFILES), default=True, action='store_bool', help='delete temporary files?', metavar='')
        operation_group.add_argument(mm(MDFORMAT), default=False, action='store_bool', help='display help in markdown format?', metavar='')
        operation_group.add_argument(mm(FORCECMD), default=False, action='store_bool', help='force cmd use (dangerous)', metavar='')

        # the repository
        repository_group = self.add_argument_group('repository arguments')
        repository_group.add_argument(mm(DATADIR), default=DEFAULT_DATADIR, action='store', help='the data directory - data, timeseries.sqlite', metavar=DIRVAR)

        # retrieval
        retrieve_group = self.add_argument_group('retrieve arguments')
        retrieve_group.add_argument(mm(TIMESPANINC), default=DEFAULT_TIMESPANINC, action='store', help='fractional increment for starting next timespan', metavar=SAMPLESVAR, type=float)
        retrieve_group.add_argument(mm(TIMESPANTOL), default=DEFAULT_TIMESPANTOL, action='store', help='fractional tolerance for overlapping timespans', metavar=SAMPLESVAR, type=float)
        retrieve_group.add_argument(mm(DOWNLOADRETRIES), default=DEFAULT_DOWNLOADRETRIES, action='store', help='maximum number of attempts to download data', metavar=NVAR, type=int)
        retrieve_group.add_argument(mm(DOWNLOADWORKERS), default=DEFAULT_DOWNLOADWORKERS, action='store', help='number of download instances to run', metavar=NVAR, type=int)
        retrieve_group.add_argument(mm(ROVERCMD), default=DEFAULT_ROVERCMD, action='store', help='command to run rover', metavar=CMDVAR)
        retrieve_group.add_argument(mm(PREINDEX), default=True, action='store_bool', help='index before retrieval?', metavar='')
        retrieve_group.add_argument(mm(INGEST), default=True, action='store_bool', help='call ingest after retrieval?', metavar='')
        retrieve_group.add_argument(mm(INDEX), default=True, action='store_bool', help='call index after ingest?', metavar='')
        retrieve_group.add_argument(mm(POSTSUMMARY), default=True, action='store_bool', help='call summary after retrieval?', metavar='')
        retrieve_group.add_argument(mm(OUTPUT_FORMAT), default=DEFAULT_OUTPUT_FORMAT, action='store', help='output data format. Choose from "mseed" (miniSEED) or "asdf" (ASDF)', metavar='')
        retrieve_group.add_argument(mm(ASDF_FILENAME), default=DEFAULT_ASDF_FILENAME, action='store', help='name of ASDF file when ASDF output is specified', metavar='')

        # metadata retrieval
        retrieve_md_group = self.add_argument_group('retrieve-metadata arguments')
        retrieve_md_group.add_argument(mm(STATIONURL), default=DEFAULT_STATIONURL, action='store', help='station service url', metavar=URLVAR)
        retrieve_md_group.add_argument(mm(FORCE_METADATA_RELOAD), default=False, action='store_bool', help='force reload of metadata', metavar='')

        # downloads
        download_group = self.add_argument_group('download arguments')
        download_group.add_argument(mm(AVAILABILITYURL), default=DEFAULT_AVAILABILITYURL, action='store', help='availability service url', metavar=URLVAR)
        download_group.add_argument(mm(DATASELECTURL), default=DEFAULT_DATASELECTURL, action='store', help='dataselect service url', metavar=URLVAR)
        download_group.add_argument(mm(TEMPDIR), default=DEFAULT_TEMPDIR, action='store', help='temporary storage for downloads', metavar=DIRVAR)
        download_group.add_argument(mm(TEMPEXPIRE), default=DEFAULT_TEMPEXPIRE, action='store', help='number of days before deleting temp files', metavar=DAYSVAR, type=int)
        download_group.add_argument(mm(HTTPTIMEOUT), default=DEFAULT_HTTPTIMEOUT, action='store', help='timeout for HTTP requests', metavar=SECSVAR, type=int)
        download_group.add_argument(mm(HTTPRETRIES), default=DEFAULT_HTTPRETRIES, action='store', help='max retries for HTTP requests', metavar=NVAR, type=int)
        download_group.add_argument(mm(FORCEFAILURES), default=DEFAULT_FORCEFAILURES, action='store', help='force failures for testing (dangerous)', metavar=PERCENTVAR, type=int)
        download_group.add_argument(mm(SORTINPYTHON), default=False, action='store_bool', help='avoid OS sort (slower)?', metavar='')

        # index
        index_group = self.add_argument_group('index arguments')
        index_group.add_argument(mm(ALL), default=False, action='store_bool', help='process all files (not just modified)?', metavar='')
        index_group.add_argument(mm(RECURSE), default=True, action='store_bool', help='when given a directory, process children?', metavar='')

        # subscription
        subscription_group = self.add_argument_group('subscription arguments')
        subscription_group.add_argument(mm(SUBSCRIPTIONSDIR), default=DEFAULT_SUBSCRIPTIONSDIR, action='store', help='directory for subscriptions', metavar=DIRVAR)
        subscription_group.add_argument(mm(RECHECKPERIOD), default=DEFAULT_RECHECKPERIOD, action='store', help='time between availabilty checks', metavar=HOURSVAR, type=int)
        subscription_group.add_argument(mm(FORCEREQUEST), default=False, action='store_bool', help='skip overlap checks (dangerous)?', metavar='')

        # logging
        logging_group = self.add_argument_group('logging arguments')
        logging_group.add_argument(mm(LOGDIR), default=DEFAULT_LOGDIR, action='store', help='directory for logs', metavar=DIRVAR)
        logging_group.add_argument(mm(LOGUNIQUE), default=False, action='store_bool', help='unique log names (with PIDs)?', metavar='')
        logging_group.add_argument(mm(LOGUNIQUEEXPIRE), default=DEFAULT_LOGUNIQUE_EXPIRE, action='store', help='number of days before deleting unique logs', metavar=DAYSVAR, type=int)
        logging_group.add_argument(mm(LOGVERBOSITY), default=DEFAULT_LOGVERBOSITY, action='store', help='log verbosity (0-6)', metavar=NVAR, type=int)
        logging_group.add_argument(mm(LOGSIZE), default=DEFAULT_LOGSIZE, action='store', help='maximum log size (e.g. 10M)', metavar=SIZE)
        logging_group.add_argument(mm(LOGCOUNT), default=DEFAULT_LOGCOUNT, action='store', help='maximum number of logs', metavar=NVAR, type=int)
        logging_group.add_argument(m(LITTLE_V), mm(VERBOSITY), default=DEFAULT_VERBOSITY, action='store', help='console verbosity (0-6)', metavar=NVAR, type=int)

        # mseedindex
        mseedindex_group = self.add_argument_group('mseedindex arguments')
        mseedindex_group.add_argument(mm(MSEEDINDEXCMD), default=DEFAULT_MSEEDINDEXCMD, action='store', help='mseedindex command', metavar=CMDVAR)
        mseedindex_group.add_argument(mm(MSEEDINDEXWORKERS), default=DEFAULT_MSEEDINDEXWORKERS, action='store', help='number of mseedindex instances to run', metavar=NVAR, type=int)

        # leap seconds
        leap_sec_group = self.add_argument_group('leap second arguments')
        leap_sec_group.add_argument(mm(LEAP), default=True, action='store_bool', help='use leap seconds file?', metavar='')
        leap_sec_group.add_argument(mm(LEAPEXPIRE), default=DEFAULT_LEAPEXPIRE, action='store', help='number of days before refreshing leap seconds file', metavar=NVAR, type=int)
        leap_sec_group.add_argument(mm(LEAPFILE), default=DEFAULT_LEAPFILE, action='store', help='file for leap second data', metavar=FILEVAR)
        leap_sec_group.add_argument(mm(LEAPURL), default=DEFAULT_LEAPURL, action='store', help='URL for leap second data', metavar=URLVAR)

        # user feedback
        user_feedback_group = self.add_argument_group('user feedback arguments')
        user_feedback_group.add_argument(mm(WEB), default=True, action='store_bool', help='auto-start the download progress web server?', metavar='')
        user_feedback_group.add_argument(mm(HTTPBINDADDRESS), default=DEFAULT_HTTPBINDADDRESS, action='store', help='bind address for HTTP server', metavar=ADDRESSVAR)
        user_feedback_group.add_argument(mm(HTTPPORT), default=DEFAULT_HTTPPORT, action='store', help='port for HTTP server', metavar=NVAR, type=int)
        user_feedback_group.add_argument(mm(EMAIL), default='', action='store', help='address for completion status', metavar=ADDRESSVAR)
        user_feedback_group.add_argument(mm(EMAILFROM), default=DEFAULT_EMAILFROM, action='store', help='from address for email', metavar=ADDRESSVAR)
        user_feedback_group.add_argument(mm(SMTPADDRESS), default=DEFAULT_SMTPADDRESS, action='store', help='address of SMTP server', metavar=ADDRESSVAR)
        user_feedback_group.add_argument(mm(SMTPPORT), default=SMTP_PORT, action='store', help='port for SMTP server', metavar=NVAR, type=int)

        # commands / args
        self.add_argument(COMMAND, metavar='COMMAND', nargs='?', type=cmd_or_alias, help='use "help" for further information')
        self.add_argument(ARGS, nargs='*', help='command arguments (depend on the command)')

    def parse_args(self, args=None, namespace=None):
        """
        Intercept normal arg parsing to:
        * scan initial args to find if config file location specified
        * if config file is missing, generate defaults
        * read config file before command line args
        If there's an error here - in particular, if the config file does not exist -
        then we continue with Nones, so that we can generate the log before raising
        an error in Config.
        """
        if args is None:
            args = sys.argv[1:]
        args = self.__preprocess_booleans(args)
        config = None
        # we run this to parse and isolate the command, but will re-run below
        # once the config file is known
        command = super().parse_args(args=args).command
        if command != INIT_REPOSITORY:
            config, args = self.__extract_config(args)
            if exists(config):
                args = self.__patch_config(args, config)

        ns = super().parse_args(args=args, namespace=namespace)

        # support rover <command> -h and rover <command> --help usage
        if mm(_help) in args or m(_h) in args:
            if command:
                ns.command = HELP_CMD
                ns.args.append(command)
    
        return ns, config

    def __preprocess_booleans(self, args):
        """
        Replace --foo with '--foo True' and --no-foo with '--foo False'.
        This makes the interface consistent with the config file (which has
        the format 'foo=True') while letting the user type simple flags.
        """
        indices = []
        for (index, arg) in enumerate(args):
            if arg.startswith(NO):
                arg = mm(arg[5:])
            for action in self._actions:
                name = mm(unbar(action.dest))
                if name == arg and type(action) is StoreBoolAction:
                    indices.append(index)
        for index in reversed(indices):
            negative = args[index].startswith(NO)
            if negative:
                args[index] = mm(args[index][5:])
            args.insert(index+1, str(not negative))
        return args

    def __extract_config(self, args):
        """
        Find the config file, if given, otherwise use the default.
        This must be done before argument parsing because we need
        to add the contents of the file to the arguments (that is
        how the file is read).
        """
        config, indices = None, []
        # find all occurences of file params, saving last
        for (index, arg) in enumerate(args):
            # must skip file name
            if arg in ('-f', '--file') and (not indices or index+1 != indices[-1]):
                if index+1 >= len(args):
                    raise Exception('No argument for %s' % arg)
                indices.append(index)
                config = args[index+1]
        # remove all occurrences
        for index in reversed(indices):
            args = args[:index] + args[index+2:]
        if not config:
            config = self.get_default(FILE)
        config = canonify(config)
        # include config flag so that it is set correctly, even if the extracted
        # value is the one that is used here
        return config, [mm(FILE), config] + args

    def write_config(self, path, args, **kargs):
        """
        If the config file is missing, fill it with values.
        If args is None, defaults are used.
        If keywords are specified, they over-ride defaults and args.
        """
        create_parents(path)

        if "WRITE_FULL_CONFIG" in kargs and not kargs["WRITE_FULL_CONFIG"]:
            actions = [action for action in self._actions if unbar(action.dest) in LITTLE_CONFIG]
        else:
            actions = self._actions

        with open(path, 'w') as out:
            for action in actions:
                name, default = action.dest, action.default
                if unbar(name) not in DYNAMIC_ARGS:
                    if default is not None:
                        if name in kargs:
                            value = kargs[name]
                        elif args is not None:  # py2.7 no __bool__ on values
                            value = getattr(args, name)
                        else:
                            value = default
                        if action.help:
                            out.write('# %s\n' % action.help)
                        out.write('%s=%s\n' % (unbar(name), value))

    def __patch_config(self, args, config):
        """
        Force the reading of the config file (ignored for init-repository).
        """
        return ['@'+config] + args

    def convert_arg_line_to_args(self, arg_line):
        '''
        Parse a line from the config file, constructing '--name value" from name=value
        and ignoring comments.
        '''
        arg_line = arg_line.strip()
        if arg_line.startswith('#'):
            return []
        elif '=' in arg_line:
            name, value = arg_line.split('=', 1)
            return [mm(name), value]
        else:
            raise Exception('Cannot parse "%s"' % arg_line)

    @staticmethod
    def __document_action(action):
        name = unbar(action.dest)
        unit = action.metavar
        default = action.default
        help = action.help
        if unit in (SECSVAR, HOURSVAR, DAYSVAR, PERCENTVAR, SAMPLESVAR):
            help = '%s (%s)' % (help, unit.lower())
        if name == FILE:
            name += ' / -f'
        elif name == HELP_CMD:
            name += ' / -h'
            default = False
            help = help.replace('this', 'the')
        elif name == FULLHELP:
            name += ' / -H'
            default = False
            help = help.replace('this', 'the')
        if help:
            help = help[0].upper() + help[1:]
        return name, default, help

    def __documentation(self, name):
        for action in self._actions:
            if name == unbar(action.dest):
                return self.__document_action(action)
        raise Exception('Unknown parameter %s' % name)

    def __documentation_names(self):
        for action in self._actions:
            name = unbar(action.dest)
            if name not in (COMMAND, ARGS):
                yield name

    NAME_WIDTH = 19
    DEFAULT_WIDTH = 20
    DESCRIPTION_WIDTH = 30

    def __print_docs_table_row(self, name, default, description):
        print('| %-*s | %-*s | %-*s |' %(self.NAME_WIDTH, name, self.DEFAULT_WIDTH, default, self.DESCRIPTION_WIDTH, description))

    def print_docs_header(self):
        """
        Print the table header row.
        """
        self.__print_docs_table_row(' Name', 'Default', 'Description')
        self.__print_docs_table_row('-' * self.NAME_WIDTH, '-' * self.DEFAULT_WIDTH, '-' * self.DESCRIPTION_WIDTH)

    def print_docs_row_md(self, name):
        """
        Print a table entry for the give argument, markdown formatted
        """
        self.__print_docs_table_row(*self.__documentation(name))

    def print_docs_row_text(self, name):
        """
        Print a table entry for the give argument, text formatted
        """
        name, default, description = self.__documentation(name)
        left = '| %-*s | %-*s' % (self.NAME_WIDTH, name, self.DEFAULT_WIDTH, default)
        right = ' | %-*s |' % (self.DESCRIPTION_WIDTH, description)
        if len(left + right) > 79:
            print('%s\n%77s |' % (left, description))
        else:
            print(left + right)

    def __print_docs_rows_md(self):
        for name in self.__documentation_names():
            self.print_docs_row_md(name)

    def print_docs_table_md(self):
        """
        Print all arguments in a table (called externally from python to generate docs).
        """
        self.print_docs_header()
        self.__print_docs_rows_md()

    def print_big_help(self, file=None):
        """
        With the new -H / --full-help option to the old implementation,
        """
        if file is None:
            file = sys.stdout
        self._print_message(self.format_big_help(), file)

    def print_help(self, file=None):
        """
        Subvert the -h and --help options to only print a restricted set of options.
        """
        if file is None:
            file = sys.stdout
        self._print_message(self.format_little_help(), file)

    def format_big_help(self):
        """
        A hacked version of format_help that prints big welcome message and
        includes all commands / actions. 
        """
        formatter = self._get_formatter()
        formatter.add_text(self.description)
        formatter.add_text(self.welcome_big_help())
        for action_group in self._action_groups:
            formatter.start_section(action_group.title)
            formatter.add_text(action_group.description)
            for action in action_group._group_actions:
                formatter.add_argument(action)
            formatter.end_section()
        formatter.add_text(self.epilog)
        return formatter.format_help()

    def format_little_help(self):
        """
        A hacked version of format_help that prints short welcome message and
        is restricted to common commands and actions in LITTLE_HELP.
        """
        formatter = self._get_formatter()
        formatter.add_text(self.description)
        formatter.add_text(self.welcome())
        for action_group in self._action_groups:
            formatter.start_section(action_group.title)
            formatter.add_text(action_group.description)
            for action in action_group._group_actions:
                if unbar(action.dest) in LITTLE_HELP:
                    formatter.add_argument(action)
            formatter.end_section()
        formatter.add_text(self.epilog)
        return formatter.format_help()

    def welcome(self):
        from rover import COMMON_COMMANDS   # avoid import loop
        return ('Usage:\n'
                '  rover <command> [args]\n\n'
                'common commands:\n'
                '{0}\n\n'
                .format(dictionary_text_list(COMMON_COMMANDS)))

    def welcome_big_help(self):
        from rover import COMMON_COMMANDS, ADVANCED_COMMANDS   # avoid import loop
        return ('Usage:\n'
                '  rover <command> [args]\n\n'
                'common commands:\n'
                '{}\n\n'                
                'advanced commands:\n'
                '{}\n\n'
                .format(dictionary_text_list(COMMON_COMMANDS),
                        dictionary_text_list(ADVANCED_COMMANDS)))


def fail_early(config):
    """
    Check commands so that we fail early.
    """
    check_cmd(config, ROVERCMD, 'rover')
    check_cmd(config, MSEEDINDEXCMD, 'mseedindex')
    workers = config.arg(DOWNLOADWORKERS)
    if workers > 10:
        raise Exception('Too many workers - risks overloading data center services (%s %d)' %
                        (mm(DOWNLOADWORKERS), workers))
    elif workers > 5:
        config.log.warn('Many workers - data center may refuse service (%s %d)' %
                        (mm(DOWNLOADWORKERS), workers))
    if config.arg(OUTPUT_FORMAT).upper() == "ASDF":
        try:
            import pyasdf
            import obspy
        except ImportError:
            raise Exception("Missing required 'pyasdf' python "
                            "package for 'output-format=asdf'")


class UserFeedback(object):
    """
    Display info on web and email status.
    """
    # can't put this in utils because we get a dependency loop.

    def __init__(self, config):
        self._log = config.log
        self._web = config.arg(WEB)
        self._url = 'http://%s:%s' % (config.arg(HTTPBINDADDRESS), config.arg(HTTPPORT))
        self._email = config.arg(EMAIL)

    def display_feedback(self):
        if self._web:
            self._log.default('Status available at %s' % self._url)
        else:
            self._log.info('No web status (see %s configuration parameter)' % mm(WEB))
        if self._email:
            self._log.default('Email status will be sent to %s' % self._email)
        else:
            self._log.info('No email status (see %s configuration parameter)' % mm(EMAIL))



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import open
from builtins import dict
from builtins import str
from future import standard_library
standard_library.install_aliases()
pkg_path = __file__

# the following replaces the default signal handler so that keyboardinterrupt
# is not raised if the user hits ctrl-C
# the value 2 matches ABORT_CODE in rover.args, but we want to do this before
# importing any other module.  see issue 46.

import os
import signal
import sys
import traceback
from collections import OrderedDict
from itertools import chain

# Set __version__ from package VERSION file
version_file = os.path.join(os.path.dirname(__file__), 'VERSION')
with open(version_file, "r") as vf:
    __version__ = vf.read().strip()


def signal_handler(sig, frame):
    sys.exit(2)

PREV_HANDLER = signal.signal(signal.SIGINT, signal_handler)


from traceback import print_exc

from .args import INIT_REPOSITORY, INDEX, INGEST, LIST_INDEX, \
    RETRIEVE, RETRIEVE_METADATA, HELP_CMD, SUBSCRIBE, DOWNLOAD, LIST_RETRIEVE, \
    START, STOP, LIST_SUBSCRIBE, UNSUBSCRIBE, DAEMON, \
    DEV, SUMMARY, LIST_SUMMARY, STATUS, WEB, TRIGGER, ABORT_CODE, ERROR_CODE
from .config import Config, RepoInitializer
from .daemon import Starter, Stopper, Daemon, StatusShower
from .download import Downloader
from .index import Indexer, IndexLister
from .ingest import Ingester
from .logs import LoggingContext
from .process import ProcessManager
from .retrieve import Retriever, ListRetriever
from .retrieve_metadata import MetadataRetriever
from .subscribe import Subscriber, SubscriptionLister, Unsubscriber, Trigger
from .summary import Summarizer, SummaryLister
from .web import ServerStarter

COMMON_COMMANDS = OrderedDict()
COMMON_COMMANDS[INIT_REPOSITORY] = (RepoInitializer, 'Create / configure the repository')
COMMON_COMMANDS[RETRIEVE] = (Retriever, 'Download, ingest and index missing data')
COMMON_COMMANDS[LIST_RETRIEVE] = (ListRetriever, 'Show what data "rover retrieve" will download')
COMMON_COMMANDS[LIST_INDEX] = (IndexLister, 'List the contents of the repository')
COMMON_COMMANDS[LIST_SUMMARY] = (SummaryLister, 'List a summary of the repository')
# so that help appears in the docs
COMMON_COMMANDS[HELP_CMD] = (None, 'Return help information about a command')

ADVANCED_COMMANDS = OrderedDict()
ADVANCED_COMMANDS[DOWNLOAD] = (Downloader, 'Download data from a remote service')
ADVANCED_COMMANDS[INDEX] = (Indexer, 'Index the repository')
ADVANCED_COMMANDS[INGEST] = (Ingester, 'Ingest data from a file into the repository')
ADVANCED_COMMANDS[WEB] = (ServerStarter, 'Start a web server showing status')
ADVANCED_COMMANDS[RETRIEVE_METADATA] = (MetadataRetriever, 'Download missing metadata')
ADVANCED_COMMANDS[SUMMARY] = (Summarizer, 'Update summary table')
#ADVANCED_COMMANDS[START] = (Starter, 'Start the background daemon')
#ADVANCED_COMMANDS[STOP] = (Stopper, 'Stop the background daemon')
#ADVANCED_COMMANDS[STATUS] = (StatusShower, 'Show the background daemon status')
#ADVANCED_COMMANDS[DAEMON] = (Daemon, 'The background daemon (prefer start/stop)')
#ADVANCED_COMMANDS[SUBSCRIBE] = (Subscriber, 'Add a subscription')
#ADVANCED_COMMANDS[LIST_SUBSCRIBE] = (SubscriptionLister, 'List the subscriptions')
#ADVANCED_COMMANDS[TRIGGER] = (Trigger, 'Ask the daemon to reprocess subscriptions')
#ADVANCED_COMMANDS[UNSUBSCRIBE] = (Unsubscriber, 'Remove subscriptions')


COMMANDS = OrderedDict(chain(COMMON_COMMANDS.items(),
                             ADVANCED_COMMANDS.items()))


def execute(command, config):
    from .help import Helper   # avoid import loop
    from .args import Arguments
    if not command:
        # print welcome message and exit
        Arguments().print_help()
        return
    commands = dict(COMMANDS)
    commands[HELP_CMD] = (Helper, '')
    if command in commands:
        commands[command][0](config).run(config.args)
    else:
        raise Exception('Unknown command %s' % command)


def main():
    config = None
    try:
        try:
            # reset the old handler now that we can catch keyboardinterrupt
            # this is necessary for windows / 2.7 where copies are not atomic and we need to catch this
            # (and it generally seems like a good idea - let's try not to change how python works too much)
            signal.signal(signal.SIGINT, PREV_HANDLER)
            config = Config()
            if config.command and config.command != HELP_CMD:
                config.lazy_validate()
            # initialise without a database...
            if not config.command or config.command in (INIT_REPOSITORY, HELP_CMD):
                execute(config.command, config)
            else:
                with ProcessManager(config):
                    execute(config.command, config)
        except KeyboardInterrupt:
            exit(ABORT_CODE)
    except Exception as e:
        if config and config.log:
            # log short error message to stdout and stack trace to a file
            with LoggingContext(config.log, handler=config.log.get_stdout_handler()):
                config.log.critical(str(e))
            with LoggingContext(config.log, handler=config.log.get_file_handler()):
                exc_type, exc_value, exc_traceback = sys.exc_info()
                error_traceback = "".join(traceback.format_exception(exc_type,
                                                                     exc_value,
                                                                     exc_traceback))
                config.log.critical(error_traceback)
            if config.command in COMMANDS:
                config.log.default('See "rover {} {}"'.format(HELP_CMD,
                                                              config.command))
            elif config.command != HELP_CMD:
                config.log.default('See "rover {}" for a list of commands'
                                   .format(HELP_CMD))
            if not config or not config._args or config.arg(DEV):
                print_exc()
        else:
            print_exc()
        # use an exit code so that workers are detected as failing
        exit(ERROR_CODE)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import str
from future import standard_library
standard_library.install_aliases()
import os
import json
import sys

from .args import DOWNLOAD, TEMPDIR, DELETEFILES, INGEST, \
    TEMPEXPIRE, HTTPTIMEOUT, \
    HTTPRETRIES, DATASELECTURL
from .ingest import Ingester
from .sqlite import SqliteSupport
from .utils import uniqueish, get_to_file, unique_filename, \
    clean_old_files, match_prefixes, create_parents, unique_path, \
    safe_unlink, post_to_file, diagnose_error

"""
The 'rover download' command - download data from a URL (and then call ingest).
"""


TMPREQUEST = 'rover_availability_request'
TMPRESPONSE = 'rover_availability_response'
TMPDOWNLOAD = 'rover_download'

# name of source when not a subscription
DEFAULT_NAME = -1


class Downloader(SqliteSupport):
    """
### Download

    rover download url [path]

    rover download file [path]

Downloads a single request, typically for a day, from a URL or a given file.
File arguments are expected to contain FDSN web services requests and fetch
data from the URL set by the dataselect-url option. Data are downloaded to a
temporary directory, which is configured by the temp-dir option.
After downloaded, data are ingested into the data-dir repository and are
deleted from the temp directory. `rover download` is called by
`rover retrieve`.

##### Significant Options

@dataselect-url
@temp-dir
@http-timeout
@http-retries
@delete-files
@ingest
@index
@verbosity
@log-dir
@log-verbosity

Options used to configure the sub-commands ingest, index are also applicable
- see Ingest/Index help for more details.


##### Examples

    rover download
    'http://service.iris.edu/fdsnws/dataselect/1/query?net=IU&sta=ANMO&loc=00&cha=BHZ&start=2010-02-27T06:30:00.000&end=2010-02-27T10:30:00.000'

will download, ingest and index data from the given URL.

    rover download myrequest.txt

will download, ingest and index data from `dataselect-url` after POSTing
`myrequest.txt`.

"""

# The only complex thing here is that these may run in parallel.  That means that
# multiple ingest instances can be running in parallel, all using mseedindex.
# To avoid conflict over sqlite access we use a different database file for each,
# so we need to track and delete those.
#
# To do this we have a table that lists ingesters, along with URLs, PIDs,
# database paths and epochs.
#
# This isn't a problem for the main ingest command because only a single instance
# of the command line command runs at any one time.

    def __init__(self, config):
        SqliteSupport.__init__(self, config)
        self._temp_dir = config.dir(TEMPDIR)
        self._dataselect_url = config.arg(DATASELECTURL)
        self._delete_files = config.arg(DELETEFILES)
        self._blocksize = 1024 * 1024
        self._ingest = config.arg(INGEST)
        self._http_timeout = config.arg(HTTPTIMEOUT)
        self._http_retries = config.arg(HTTPRETRIES)
        self._config = config
        clean_old_files(self._temp_dir, config.arg(TEMPEXPIRE), match_prefixes(TMPDOWNLOAD), self._log)

    def run(self, args):
        """
        Download the give URL or file, then call ingest and index before deleting.
        """
        if len(args) < 1 or len(args) > 2:
            raise Exception('Usage: rover %s url [path]' % DOWNLOAD)
        in_path_or_url = args[0]
        db_path = self._ingesters_db_path(in_path_or_url)
        if '://' in in_path_or_url:
            url, in_path, get = in_path_or_url, None, True
            if '&' not in url:
                self._log.warn(('The URL provided is "%s" - this does not contain any ampersands.  ' +
                                'You may have forgotten to quote the URL (and should expect failure).') % url)
        else:
            url, in_path, get = self._dataselect_url, in_path_or_url, False
            if not os.path.exists(in_path):
                raise Exception('Could not find file "%s"' % in_path)
        if len(args) == 2:
            out_path, delete_out = args[2], False
        else:
            out_path, delete_out = unique_path(self._temp_dir, TMPDOWNLOAD, in_path_or_url), True

        try:
            if self._do_download(get, url, in_path, out_path):  # False when no data available
                if self._ingest:
                    Ingester(self._config).run([out_path], db_path=db_path)
        except:
            raise
        else:
            if self._delete_files:
                # Remove empty log files to avoid clutter
                log_path = self._config.log_path
                if os.path.exists(log_path) and os.path.getsize(log_path) == 0:
                    safe_unlink(log_path)
        finally:
            if self._delete_files:
                if delete_out:
                    safe_unlink(out_path)
                safe_unlink(db_path)

    def _do_download(self, get, url, in_path, out_path):
        # previously we extracted the file name from the header, but the code
        # failed in python 2 (looked like a backport library bug), so now we let the user specify,
        if os.path.exists(out_path):
            raise Exception('Path %s for download already exists' % out_path)
        create_parents(out_path)

        if get:
            response, check_status = get_to_file(url, out_path,
                                                 self._http_timeout, self._http_retries, self._log)
            check_status()
        else:
            try:
                response, check_status = post_to_file(url, in_path, out_path,
                                                      self._http_timeout, self._http_retries, self._log)
                check_status()
            except Exception as e:
                diagnose_error(self._log, str(e), in_path, out_path, copied=False)
                raise

        # write feedback, in JSON, to caller on stdout with download byte count
        if response and os.path.isfile(response):
            feedback = {}
            feedback['download_byte_count'] = os.path.getsize(response)

            if feedback['download_byte_count']:
                sys.stdout.write(json.dumps(feedback))

        return response

    def _ingesters_db_path(self, url):
        name = uniqueish('rover_ingester', url)
        return unique_filename(os.path.join(self._temp_dir, name))



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import super
from builtins import object
from future import standard_library
standard_library.install_aliases()
from sqlite3 import connect

from .utils import canonify


"""
Support for database access.
"""


def init_db(dbpath, log):
    """
    Open a connection to the database.
    """

    log.debug('Connecting to sqlite3 %s' % dbpath)
    db = connect(dbpath, timeout=60.0)
    # https://www.sqlite.org/foreignkeys.html
    db.execute('PRAGMA foreign_keys = ON')
    db.execute('PRAGMA case_sensitive_like = ON')  # as used by mseedindex
    return db


class NoResult(Exception):
    """
    Exception thrown when no results available.
    """

    def __init__(self, sql, params):
        super().__init__('%s %s' % (sql, params))


class CursorContext(object):
    """
    Create a cursor for the duration of the scope (and then close it).
    """

    def __init__(self, support, quiet):
        self._support = support
        self._quiet = quiet
        self._cursor = support._db.cursor()

    def __enter__(self):
        return self._cursor

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            if self._quiet:
                self._support._log.debug('Cursor exit: %s' % exc_val)
            else:
                self._support._log.error('Cursor exit: %s' % exc_val)
        else:
            self._support._db.commit()  # probably implied by close?
        self._cursor.close()
        return False  # propagate any exceptions


class SqliteDb(object):
    """
    Base class with utilities for accessing database.
    """

    def __init__(self, db, log):
        self._db = db
        self._log = log

    def cursor(self, quiet=False):
        return CursorContext(self, quiet)

    def execute(self, sql, params=tuple(), quiet=False):
        """
        Execute a single command in a transaction.
        """
        with self.cursor(quiet=quiet) as c:
            self._log.debug('Execute: %s %s' % (sql, params))
            c.execute(sql, params)

    def fetchsingle(self, sql, params=tuple(), quiet=False):
        """
        Return a single value from a select.

        Raise NoResult if no value.
        """
        with self.cursor(quiet=quiet) as c:
            self._log.debug('Fetchsingle: %s %s' % (sql, params))
            result = c.execute(sql, params).fetchone()
            if result:
                if len(result) > 1:
                    raise Exception('Multiple results for "%s %s"' % (sql, params))
                else:
                    return result[0]
            else:
                raise NoResult(sql, params)

    def fetchone(self, sql, params=tuple(), quiet=False):
        """
        Return a single row from a select.

        Raise NoResult if no row.
        """
        with self.cursor(quiet=quiet) as c:
            self._log.debug('Fetchone: %s %s' % (sql, params))
            result = c.execute(sql, params).fetchone()
            if result:
                return result
            else:
                raise NoResult(sql, params)

    def fetchall(self, sql, params=tuple(), quiet=False):
        """
        Return a list of rows from a select.

        Note that it may be better (eg in list-index) to iterate over
        the cursor explicitly (see foreachrow).
        """
        with self.cursor() as c:
            self._log.debug('Fetchall: %s %s' % (sql, params))
            return c.execute(sql, params).fetchall()

    def foreachrow(self, sql, params, callback, quiet=False):
        """
        Call the callback for each row in the results.
        """
        with self.cursor(quiet=quiet) as c:
            self._log.debug('foreachrow: %s %s' % (sql, params))
            for row in c.execute(sql, params):
                callback(row)

    def close(self):
        self._db.close()


class SqliteContext(object):
    """
    Create a database connection for the duration of the scope.
    """

    def __init__(self, file, log):
        self._db = SqliteDb(connect(file, timeout=60.0), log)

    def __enter__(self):
        return self._db

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._db.close()
        return False  # propagate any exceptions


class SqliteSupport(SqliteDb):
    """
    Alternative constructor for SqliteDb.
    """

    def __init__(self, config):
        super().__init__(config.db, config.log)
